/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 19.04.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.knime.core.data.DataCell;
import org.knime.core.data.DataColumnSpec;
import org.knime.core.data.DataColumnSpecCreator;
import org.knime.core.data.def.DoubleCell;
import org.knime.core.data.def.IntCell;
import org.knime.core.node.NodeLogger;
import org.knime.network.core.api.GraphObjectIterator;
import org.knime.network.core.api.KPartiteGraph;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.GraphFactory;
import org.knime.network.core.core.PartitionType;
import org.knime.network.core.core.exception.InvalidFeatureException;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.feature.FeatureTypeFactory;
import org.knime.network.core.core.util.GraphUtil;
import org.knime.network.core.knime.NetCol;

/**
 *
 * @author Tobias Koetter
 */
public class ConceptGraph {
    private static final NodeLogger LOGGER = NodeLogger.getLogger(ConceptGraph.class);
    /**Members node partition name.*/
    public static final String MEMBERS_PARTITION =  "members";
    /**Aspects node partition name.*/
    public static final String ASPECTS_PARTITION =  "aspects";
    /**Concepts node partition name.*/
    public static final String CONCEPTS_PARTITION =  "concepts";
    /**Cue validity feature.*/
    public static final String CUE_VALIDITY_FEATURE = "cueValidity";
    /**Density feature.*/
    public static final String DENSITY_FEATURE = "density";
    /**Score feature.*/
    public static final String SCORE_FEATURE = "score";

    /**
     * Provides the successors of a given node.
     * @author koetter
     */
    interface NeighborProvider {
        /**
         * @param memberId the object id
         * @return the {@link Set} of its neighbors
         * @throws PersistenceException if an exception occurs
         */
        Set<String> getPotentialAspects(String memberId) throws PersistenceException;
    }

    private KPartiteGraph<PersistentObject, Partition> m_net = null;
    private final List<ConceptGraphObject> m_concepts;
    private final List<ConceptGraphObject> m_aspects;
    private final List<ConceptGraphObject> m_members;
    private final String m_id;
    private final ConceptGraphScores m_memberScore;
    private final ConceptGraphScores m_aspectScore;
    private final ConceptGraphScores m_conceptScore;
    private final int m_noOfEdges;
    private final double m_totalScore;
    private final PersistentObject m_startNode;
    private final int m_noOfPossibleMembers;
    private final int m_noOfPossibleAspects;

    /**
     * @param view the original {@link KPartiteGraphView}
     * @param aspects the aspects
     * @param members the members
     * @param score the score of the concept graph
     * @param startNode the start node
     * @param noOfPossibleAspects number of possible aspects
     * @param noOfPossibleMembers number of possible members
     */
    public ConceptGraph(final KPartiteGraphView<PersistentObject, Partition> view,
                        final Collection<ConceptGraphObject> aspects,  final Collection<ConceptGraphObject> members,
                        final double score, final PersistentObject startNode, final int noOfPossibleAspects,
                        final int noOfPossibleMembers) {
        m_startNode = startNode;
        m_noOfPossibleMembers = noOfPossibleMembers;
        m_noOfPossibleAspects = noOfPossibleAspects;
//        final long startTime = System.currentTimeMillis();
        m_totalScore = score;
        m_concepts = new LinkedList<ConceptGraphObject>();
        m_conceptScore = new ConceptGraphScores();
        m_aspects = new LinkedList<ConceptGraphObject>();
        m_aspectScore = new ConceptGraphScores();
        m_memberScore = new ConceptGraphScores();
        double conceptCueValidity = -1;
        for (final ConceptGraphObject aspect : aspects) {
            final double cueValidity = aspect.getCueValidity();
//            final boolean isStartNode = m_startNodeIds.contains(aspect.getId());
//            final boolean isConceptCandidate = isStartNode;
            boolean isConceptCandidate = isConceptCandidate(aspect, members);
            if (cueValidity > conceptCueValidity) {
                if (isConceptCandidate) {
                    for (final ConceptGraphObject concept : m_concepts) {
                        addAspect(concept, members.size());
                    }
                    resetConcepts();
                    addConcept(aspect, members.size());
                    conceptCueValidity = cueValidity;
                }
            } else if (cueValidity == conceptCueValidity) {
//                isConceptCandidate = isConceptCandidate(aspect, members);
                if (isConceptCandidate) {
                    addConcept(aspect, members.size());
                }
            } else {
                isConceptCandidate = false;
            }
            if (!isConceptCandidate) {
                addAspect(aspect, members.size());
            }
        }
        int edgeCounter = 0;
        m_members = new ArrayList<ConceptGraphObject>(members.size());
        for (final ConceptGraphObject member : members) {
            addMember(member, aspects.size());
            edgeCounter += member.getInDegree();
        }
        //sort the aspects and members based on their cue validity
        Collections.sort(m_aspects, ConceptGraphObject.DESC_CV_COMP);
        Collections.sort(m_members, ConceptGraphObject.DESC_CV_COMP);
        final String conceptBasedId = GraphUtil.convertToString(null, null, ", ",
            Arrays.asList(ConceptGraphObject.createIdArray(m_concepts, true)));
        if (conceptBasedId != null && !conceptBasedId.trim().isEmpty()) {
            m_id = conceptBasedId;
        } else {
            m_id = "noConcept_" + System.currentTimeMillis();
        }
        m_noOfEdges = edgeCounter;
//        final long endTime = System.currentTimeMillis();
//        final long durationTime = endTime - startTime;
//        LOGGER.info(m_id + ": Time for concept graph creation: " + durationTime + " ms");
//        LOGGER.info(m_id + ": No of concepts: " + m_concepts.size() + " no of aspects: " + m_aspects.size()
//            + " no of members: " + m_members.size());
    }

    /**
     *
     */
    private void resetConcepts() {
        m_concepts.clear();
        m_conceptScore.reset();
    }

    /**
     * @param member
     * @param size
     * @return
     */
    private boolean addMember(final ConceptGraphObject member, final int size) {
        m_memberScore.add(member, size);
        return m_members.add(member);
    }

    /**
     * @param aspect
     * @param size
     */
    private void addAspect(final ConceptGraphObject aspect, final int size) {
        m_aspectScore.add(aspect, size);
        m_aspects.add(aspect);
    }

    /**
     * @param concept
     * @param size
     */
    private void addConcept(final ConceptGraphObject concept, final int size) {
        m_conceptScore.add(concept, size);
        m_concepts.add(concept);
    }

    private boolean isConceptCandidate(final ConceptGraphObject aspect, final Collection<ConceptGraphObject> members) {
        return aspect.getInDegree() >= members.size();
    }

    private static PersistentObject createNode(final KPartiteGraph<PersistentObject, Partition> net,
                        final ConceptGraphObject object, final Partition partition, final int size,
                        final ConceptGraphScores scorer)
                                throws PersistenceException, InvalidFeatureException {
        final PersistentObject node = net.createObject(object.getId());
        net.assignToPartition(node, partition);
        final double cueValidity = object.getCueValidity();
        net.addFeature(node, CUE_VALIDITY_FEATURE, cueValidity);
        final double density = object.getOverlap(size);
        net.addFeature(node, DENSITY_FEATURE, density);
        final double score = object.getScore();
        net.addFeature(node, SCORE_FEATURE, score);
        return node;
    }

    /**
     * @param view the original {@link KPartiteGraphView}
     * @return the {@link KPartiteGraphView} representing the concept graph
     * @throws PersistenceException if an exception occurs
     */
    public KPartiteGraphView<PersistentObject, Partition> getConceptGraph(
                              final KPartiteGraphView<PersistentObject, Partition> view) throws PersistenceException {
        if (m_net == null) {
            m_net = createConceptGraph(new NeighborProvider() {

                @Override
                public Set<String> getPotentialAspects(final String id) throws PersistenceException {
                    return getAdjacentNodeSet(view, id, true);
                }
            });
        }
        return m_net;
    }

    /**
     * @param neighborMap {@link Map} with {@link NodeNeighbors}s
     * @return the {@link KPartiteGraphView} representing the concept graph
     * @throws PersistenceException if an exception occurs
     */
    public KPartiteGraphView<PersistentObject, Partition> getConceptGraph(final Map<String, NodeNeighbors> neighborMap)
            throws PersistenceException {
        if (m_net == null) {
            m_net = createConceptGraph(new NeighborProvider() {
                @Override
                public Set<String> getPotentialAspects(final String id) {
                    final NodeNeighbors neighbor = neighborMap.get(id);
                    return neighbor.getSuccessors();
                }
            });
        }
        return m_net;
    }

    private KPartiteGraph<PersistentObject, Partition> createConceptGraph(final NeighborProvider np)
            throws PersistenceException {
        final KPartiteGraph<PersistentObject, Partition> net = GraphFactory.createNet(m_id);
        final Partition memberPartition = net.createPartition(ConceptGraph.MEMBERS_PARTITION, PartitionType.NODE);
        final Partition aspectPartition = net.createPartition(ConceptGraph.ASPECTS_PARTITION, PartitionType.NODE);
        final Partition conceptPartition = net.createPartition(ConceptGraph.CONCEPTS_PARTITION, PartitionType.NODE);
        try {
            net.defineFeature(FeatureTypeFactory.getDoubleType(), ConceptGraph.CUE_VALIDITY_FEATURE);
            net.defineFeature(FeatureTypeFactory.getDoubleType(), ConceptGraph.DENSITY_FEATURE);
            net.defineFeature(FeatureTypeFactory.getDoubleType(), ConceptGraph.SCORE_FEATURE);
            final Map<String, PersistentObject> aspectMap = new HashMap<String, PersistentObject>();
            for (final ConceptGraphObject object : m_concepts) {
                final PersistentObject node =
                        createNode(net, object, conceptPartition, getNoOfMembers(), m_conceptScore);
                aspectMap.put(object.getId(), node);
            }
            for (final ConceptGraphObject aspect : m_aspects) {
                final PersistentObject aspectNode =
                        createNode(net, aspect, aspectPartition, getNoOfMembers(), m_aspectScore);
                aspectMap.put(aspect.getId(), aspectNode);
            }
            for (final ConceptGraphObject member : m_members) {
                final PersistentObject memberNode =
                        createNode(net, member, memberPartition, getNoOfMembers(), m_memberScore);
                final Set<String> potentialAspects = np.getPotentialAspects(member.getId());
                for (final Entry<String, PersistentObject> aspectEntry : aspectMap.entrySet()) {
                    if (potentialAspects.contains(aspectEntry.getKey())) {
                    final PersistentObject aspect = aspectEntry.getValue();
                        final String edgeId = aspect.getId() + "->" + memberNode.getId();
                        net.createDirectedEdge(edgeId, memberNode, aspect);
                    }
                }
            }
            net.commit();
            return net;
        } catch (final InvalidFeatureException e) {
            //this shouldn't happen
            throw new PersistenceException(e);
        }
    }

    /**
     * @param view the {@link KPartiteGraphView}
     * @param nodeId the id of the node to get the neighbors for
     * @param isSource <code>true</code> if the node is the source and its target neighbors should be extracted
     * @return the set of neighbors
     * @throws PersistenceException if an exception occurs
     */
    public static Set<String> getAdjacentNodeSet(final KPartiteGraphView<PersistentObject, Partition> view,
                                           final String nodeId, final boolean isSource) throws PersistenceException {
        final Collection<Partition> nodePartitions = view.getPartitions(PartitionType.NODE);
        final Collection<Partition> edgePartitions = view.getPartitions(PartitionType.EDGE);
        GraphObjectIterator<PersistentObject> neighbors = null;
        GraphObjectIterator<PersistentObject> incidentEdges = null;
        final HashSet<String> set = new HashSet<String>();
        try {
            final PersistentObject node = view.getObject(nodeId);
            incidentEdges = view.getIncidentObjects(node, edgePartitions);
            for (final PersistentObject edge : incidentEdges) {
                if (isSource && !view.isSource(edge, node) || !isSource && !view.isTarget(edge, node)) {
                    continue;
                }
                neighbors = view.getIncidentObjects(edge, nodePartitions);
                for (final PersistentObject neighbor : neighbors) {
                    if (isSource && !view.isTarget(edge, neighbor)
                            || !isSource && !view.isSource(edge, neighbor)) {
                        continue;
                    }
                    set.add(neighbor.getId());
                }
            }
            return set;
        } finally {
            GraphUtil.close(neighbors);
            GraphUtil.close(incidentEdges);
        }
    }

    /**
     * @return the start node
     */
    public PersistentObject getStartNode() {
        return m_startNode;
    }

    /**
     * @return the concept nodes
     */
    public String[] getConceptNodes() {
        return ConceptGraphObject.createIdArray(m_concepts, true);
    }

    /**
     * @return the aspect nodes
     */
    public String[] getAspectNodes() {
        return ConceptGraphObject.createIdArray(m_aspects, false);
    }

    /**
     * @return the member nodes
     */
    public String[] getMemberNodes() {
        return ConceptGraphObject.createIdArray(m_members, false);
    }

    /**
     * @return the member {@link ConceptGraphScores}
     */
    public ConceptGraphScores getMemberScores() {
        return m_memberScore;
    }

    /**
     * @return the concept {@link ConceptGraphScores}
     */
    public ConceptGraphScores getConceptScores() {
        return m_conceptScore;
    }

    /**
     * @return the aspect {@link ConceptGraphScores}
     */
    public ConceptGraphScores getAspectScores() {
        return m_aspectScore;
    }

    /**
     * @return the size score
     */
    public double getSizeScore() {
        return getNoOfEdges() * getNoOfNodes();
    }

    /**
     * @return the total score
     */
    public double getTotalScore() {
        return m_totalScore;
    }

    /**
     * @return the quality score
     */
    public double getQualityScore() {
        final int noOfAspects = getNoOfConcepts() + getNoOfAspects();
        final double maxConceptCV = getConceptScores().getMaximum(ConceptGraphScores.CUE_VALIDITY);
        return getDensity() * noOfAspects * getNoOfMembers() * maxConceptCV;
    }

    /**
     * @return the basic concept score
     */
    public double getBasicConceptScores() {
        return getConceptScores().getMaximum(ConceptGraphScores.CUE_VALIDITY)
                * getAspectScores().getMean(ConceptGraphScores.CUE_VALIDITY);
    }

    /**
     * @return the number of edges
     */
    public int getNoOfEdges() {
        return m_noOfEdges;
    }

    /**
     * @return the number of nodes
     */
    public int getNoOfNodes() {
        return getNoOfConcepts() + getNoOfAspects() + getNoOfMembers();
    }

    /**
     * @return number of aspects
     */
    public int getNoOfAspects() {
        return m_aspects.size();
    }

    /**
     * @return the number of members
     */
    public int getNoOfMembers() {
        return m_members.size();
    }

    /**
     * @return the number of concepts
     */
    public int getNoOfConcepts() {
        return m_concepts.size();
    }


    /**
     * @return the maximum number of edges between members and their aspects and the concepts
     */
    public int getMaxNoOfEdges() {
        return (getNoOfConcepts() + getNoOfAspects()) * getNoOfMembers();
    }

    /**
     * @return the noOfPossibleAspects
     */
    public int getNoOfPossibleAspects() {
        return m_noOfPossibleAspects;
    }

    /**
     * @return the noOfPossibleMembers
     */
    public int getNoOfPossibleMembers() {
        return m_noOfPossibleMembers;
    }

    /**
     * @return the edge density
     */
    public double getDensity() {
        final double maxNoOfEdges = getMaxNoOfEdges();
        if (maxNoOfEdges == 0) {
            return 0;
        }
        return getNoOfEdges() / maxNoOfEdges;
    }

    /**
     * @return the score {@link DataCell}s
     */
    public Collection<? extends DataCell> getScoreCells() {
        final Collection<DataCell> cells = new LinkedList<DataCell>();
        cells.add(new DoubleCell(getTotalScore()));
        cells.add(new DoubleCell(getQualityScore()));
        cells.add(new DoubleCell(getDensity()));
        cells.add(new DoubleCell(getSizeScore()));
        cells.add(new DoubleCell(getBasicConceptScores()));
        cells.add(new IntCell(getNoOfNodes()));
        cells.add(new IntCell(getNoOfConcepts()));
        cells.add(new IntCell(getNoOfAspects()));
        cells.add(new IntCell(getNoOfMembers()));
        cells.add(new IntCell(getNoOfEdges()));
        cells.add(new IntCell(getMaxNoOfEdges()));
        cells.add(new IntCell(getNoOfPossibleAspects()));
        cells.add(new IntCell(getNoOfPossibleMembers()));
        cells.add(new DoubleCell(m_conceptScore.getMaximum(ConceptGraphScores.CUE_VALIDITY)));
        cells.add(new DoubleCell(m_conceptScore.getMaximum(ConceptGraphScores.SCORE)));
        addStatCells(cells, m_aspectScore, ConceptGraphScores.CUE_VALIDITY, ConceptGraphScores.DENSITY,
                     ConceptGraphScores.SCORE);
        addStatCells(cells, m_memberScore, ConceptGraphScores.CUE_VALIDITY, ConceptGraphScores.DENSITY,
                     ConceptGraphScores.SCORE);
        return cells;
    }

    /**
     * @return the score specifications
     */
    public static Collection<? extends DataColumnSpec> createScoreSpecs() {
        final Collection<DataColumnSpec> specs = new LinkedList<DataColumnSpec>();
        final DataColumnSpecCreator creator = new DataColumnSpecCreator("Dummy", IntCell.TYPE);
        creator.setType(DoubleCell.TYPE);
        creator.setName("Total score");
        specs.add(creator.createSpec());
        creator.setName("Quality score");
        specs.add(creator.createSpec());
        creator.setName("Edge density");
        specs.add(creator.createSpec());
        creator.setName("Size score");
        specs.add(creator.createSpec());
        creator.setName("Basic concept score");
        specs.add(creator.createSpec());
        creator.setType(IntCell.TYPE);
        creator.setName(NetCol.NO_OF_NODES);
        specs.add(creator.createSpec());
        creator.setName("No of concepts");
        specs.add(creator.createSpec());
        creator.setName("No of aspects");
        specs.add(creator.createSpec());
        creator.setName("No of members");
        specs.add(creator.createSpec());
        creator.setName(NetCol.NO_OF_EDGES);
        specs.add(creator.createSpec());
        creator.setName("Max " + NetCol.NO_OF_EDGES);
        specs.add(creator.createSpec());
        creator.setName("No of possible aspects");
        specs.add(creator.createSpec());
        creator.setName("No of possible members");
        specs.add(creator.createSpec());
        creator.setType(DoubleCell.TYPE);
        creator.setName("Concept cue validity");
        specs.add(creator.createSpec());
        creator.setName("Concept score");
        specs.add(creator.createSpec());
        addStatSpecs(specs, creator, "aspect", "cue validity", "density", "score");
        addStatSpecs(specs, creator, "member", "cue validity", "density", "score");
        return specs;
    }

    private void addStatCells(final Collection<DataCell> cells, final ConceptGraphScores scorer, final int... idx) {
        for (final int i : idx) {
            cells.add(new DoubleCell(scorer.getMinimum(i)));
            cells.add(new DoubleCell(scorer.getMaximum(i)));
            cells.add(new DoubleCell(scorer.getMean(i)));
            cells.add(new DoubleCell(scorer.getVariance(i)));
        }
    }

    private static void addStatSpecs(final Collection<DataColumnSpec> specs, final DataColumnSpecCreator creator,
                                     final String type, final String... scores) {
        creator.setType(DoubleCell.TYPE);
        for (final String score : scores) {
            creator.setName("Min. " + type + " " + score);
            specs.add(creator.createSpec());
            creator.setName("Max. " + type + " " + score);
            specs.add(creator.createSpec());
            creator.setName("Mean " + type + " " + score);
            specs.add(creator.createSpec());
            creator.setName("Variance " + type + " " + score);
            specs.add(creator.createSpec());
        }
    }

    /**
     * @return the id of the network based on the start nodes
     */
    public String getId() {
        return m_id;
    }
}
