/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 14.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.NodeNeighbors;
import java.util.Map;
import org.knime.core.node.util.ButtonGroupEnumInterface;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;

/**
 *
 * @author Tobias Koetter
 */
public enum ConceptGraphExtractorType implements ButtonGroupEnumInterface {
    /**Aspect based extractor.*/
    ASPECT("Aspect", false),
    /**Cue validity aware aspect based extractor.*/
    ASPECT_CUE_VALIDITY("Aspect(cue validity)", false),
    /**Iterative extractor.*/
    ITERATIVE("Iteractive aspect based", true),
    /**Iterative extractor.*/
    ITERATIVE_ONLY("Iteractive", false);

    private final String m_title;
    private final boolean m_isDefault;

    private ConceptGraphExtractorType(final String title, final boolean isDefault) {
        m_title = title;
        m_isDefault = isDefault;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getText() {
        return m_title;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActionCommand() {
        return name();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getToolTip() {
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public boolean isDefault() {
        return m_isDefault;
    }

    /**
     * @param action the action of the scorer
     * @return the {@link ConceptGraphExtractorType} with the given action
     */
    public static ConceptGraphExtractorType get(final String action) {
        return valueOf(action);
    }

    /**
     * @param view the {@link KPartiteGraphView}
     * @param scorer the {@link Scorer} to use
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the starting node
     * @return the {@link ConceptGraphExtractor}
     */
    public ConceptGraphExtractor getExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                                       final Scorer scorer, final int minNodeDegree,
                                       final Map<String, NodeNeighbors> neighborMap, final PersistentObject startNode) {
        switch (this) {
            case ASPECT:
                return new FastAspectBasedConceptGraphExtractor(view, scorer, minNodeDegree, neighborMap, startNode);
            case ASPECT_CUE_VALIDITY:
                return new CueValidityAwareAspectBasedConceptGraphExtractor(view, scorer, minNodeDegree,
                    neighborMap, startNode);
            case ITERATIVE:
                return new IterativeFastAspectBasedConceptGraphExtractor(view, scorer, minNodeDegree, neighborMap,
                    startNode);
            case ITERATIVE_ONLY:
                return new IterativeConceptGraphExtractor(view, scorer, minNodeDegree, neighborMap, startNode);
        }
        throw new RuntimeException("Invalid ConceptGraph extracor");
    }

    /**
     * @return the default type to use
     */
    public static ConceptGraphExtractorType getDefault() {
        for (final ConceptGraphExtractorType type : values()) {
            if (type.isDefault()) {
                return type;
            }
        }
        return values()[0];
    }
}
