/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 28.04.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import java.util.Arrays;

/**
 *
 * @author Tobias Koetter
 */
public class ConceptGraphScores {
    /***Cue validity index: {@value}.*/
    public static final int CUE_VALIDITY = 0;
    /***Density index: {@value}.*/
    public static final int DENSITY = 1;
    /***Score index: {@value}.*/
    public static final int SCORE = 2;
    private static final int NO_OF_STATS = 3;

    private int m_count = 0;
    private final double[] m_mean = new double[NO_OF_STATS];
    private final double[] m_min = new double[NO_OF_STATS];
    private final double[] m_max = new double[NO_OF_STATS];
    private final double[] m_sum = new double[NO_OF_STATS];
    private final double[] m_squareSum = new double[NO_OF_STATS];

    /**
     * Constructor.
     */
    public ConceptGraphScores() {
        reset();
    }

    /**
     * Resets the scorer.
     */
    public void reset() {
        Arrays.fill(m_min, Double.POSITIVE_INFINITY);
        Arrays.fill(m_max, Double.NEGATIVE_INFINITY);
        Arrays.fill(m_sum, 0);
        Arrays.fill(m_squareSum, 0);
    }
    /**
     * @param object the object to add
     * @param size the size of the neighbor set
     */
    public void add(final ConceptGraphObject object, final int size) {
        final double cueValidity = object.getCueValidity();
        addValue(cueValidity, CUE_VALIDITY);
        final double density = object.getOverlap(size);
        addValue(density, DENSITY);
        final double score = object.getScore();
        addValue(score, SCORE);
        m_count++;
    }

    private void addValue(final double value, final int idx) {
        m_mean[idx] = m_mean[idx] * ((double)m_count / (m_count + 1))
                + value * (1.0 / (m_count + 1));
        if (value < m_min[idx]) {
            m_min[idx] = value;
        }
        if (value > m_max[idx]) {
            m_max[idx] = value;
        }
        m_sum[idx] += value;
        m_squareSum[idx] += value * value;
    }

    /**
     * @param idx the index of the score of interest
     * @return the variance of the score
     */
    public double getVariance(final int idx) {
        if (m_count == 1) {
            return 0;
        }
        double variance = (m_squareSum[idx] - ((m_sum[idx] * m_sum[idx])
                / m_count)) / (m_count - 1);
        // unreported bug fix: in cases in which a column contains
        // almost only one value (for instance 1.0) but one single
        // 'outlier' whose value is, for instance 0.9999998, we get
        // round-off errors resulting in negative variance values
        if (variance < 0.0 && variance > -1.0E8) {
            variance = 0.0;
        }
        return variance;
    }

    /**
     * @param idx the index of the score of interest
     * @return the minimum score
     */
    public double getMinimum(final int idx) {
        return m_min[idx];
    }

    /**
     * @param idx the index of the score of interest
     * @return the maximum score
     */
    public double getMaximum(final int idx) {
        return m_max[idx];
    }

    /**
     * @param idx the index of the score of interest
     * @return the mean score
     */
    public double getMean(final int idx) {
        return m_mean[idx];
    }

    /**
     * @return the preferred score
     */
    public double getScore() {
        return getMinimum(SCORE);
    }
}