/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 07.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.AspectComparator;
import edu.cmu.koetter.util.NeighborBiGraphExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.NodeLogger;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;

/**
 *
 * @author Tobias Koetter
 */
public class CueValidityAwareAspectBasedConceptGraphExtractor extends FastConceptGraphExtractor {

    private static final NodeLogger LOGGER =
            NodeLogger.getLogger(CueValidityAwareAspectBasedConceptGraphExtractor.class);

    /**
     * @param view the {@link KPartiteGraphView}
     * @param objectScorer the {@link Scorer}
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the start node
     */
    public CueValidityAwareAspectBasedConceptGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                            final Scorer objectScorer, final int minNodeDegree,
                            final Map<String, NodeNeighbors> neighborMap, final PersistentObject startNode) {
        super(view, objectScorer, minNodeDegree, neighborMap, startNode);
    }

    /**
     * @param exec {@link ExecutionMonitor} to provide progress
     * @return the extracted {@link ConceptGraph}
     * @throws PersistenceException if an exception occurs
     * @throws CanceledExecutionException if the operation has been canceled
     */
    @Override
    protected Collection<ConceptGraph> findConceptGraph(final ExecutionMonitor exec) throws PersistenceException,
            CanceledExecutionException {
        exec.setMessage("analyzing aspects and members...");
        final NeighborBiGraphExtractor extractor = getExtractor();
        final Map<String, Set<String>> aspectNeighbors = extractor.getAspectNeighbors();
        final Map<String, Set<String>> memberNeighbors = extractor.getMemberNeighbors();
        final Map<String, ConceptGraphObject> memberMap =
                createObjectMap(memberNeighbors, aspectNeighbors.keySet(), false);
        final Map<String, ConceptGraphObject> aspectMap =
                createObjectMap(aspectNeighbors, memberNeighbors.keySet(), true);
        final List<ConceptGraphObject> sortedMembers = new ArrayList<ConceptGraphObject>(memberMap.values());
        //the order of the members has no influence on the score
//        Collections.sort(sortedMembers, new Comparator<ConceptGraphObject>() {
//          @Override
//          public int compare(final ConceptGraphObject o1, final ConceptGraphObject o2) {
//              return o1.getInDegree() - o2.getInDegree();
//          }
//        });
        //sort the aspects descending based on the density and cue validity with respect to all possible members
        final List<ConceptGraphObject> sortedAspects = new ArrayList<ConceptGraphObject>(aspectMap.values());
        //compute the score for all aspects based on all members
        Collections.sort(sortedAspects, AspectComparator.getInstance());
        double maxScore = 0;
        final Collection<ConceptGraphObject> members = new ArrayList<ConceptGraphObject>();
        final Collection<ConceptGraphObject> aspects = new ArrayList<ConceptGraphObject>();
        while (!sortedAspects.isEmpty() && !memberMap.isEmpty()) {
            exec.checkCanceled();
          final ScoreCache memberScores = new ScoreCache(memberMap.size(), null);
          double conceptGraphScore = 0;
          final int[] aspectInDegrees = new int[sortedAspects.size()];
          final BitSet aNeighborIdxs = new BitSet(sortedAspects.size());
          int idx = 0;
          Arrays.fill(aspectInDegrees, 0);
          for (ConceptGraphObject member : sortedMembers) {
              double cueValDiff = 0;
              double mScore = 0;
              idx = 0;
              aNeighborIdxs.clear();
              for (ConceptGraphObject aspect : sortedAspects) {
                  final Set<String> aNeighbors = aspectNeighbors.get(aspect.getId());
                  final double aNeighborCount = aNeighbors.size();
                  final int aInDegree = aspectInDegrees[idx];
                  if (aNeighbors.contains(member.getId())) {
                      //the member is a neighbor of the aspect
                      mScore += (aInDegree + 1) / aNeighborCount;
                      //this is the cue validity diff of all members that possess the aspect minus the members
                      //that do not possess the aspect
                      cueValDiff += (aInDegree - (memberScores.size() - aInDegree)) * 1 / aNeighborCount;
                      aNeighborIdxs.set(idx);
                  } else {
                      //the aspect is no member of the neighbor so we do not need to compute the cue validity diff
                      mScore -= aInDegree / aNeighborCount;
                      //cueValDiff += 0; size the cue validity changes only if the aspect is a neighbor of the member
                  }
                  //increase the aspect in degree index prior processing the next index
                  idx++;
              }
              final double memberScore = mScore + cueValDiff;
              final boolean add = memberScore >= 0;
              //loop over all aspects that are affected by the current member
              for (int i = aNeighborIdxs.nextSetBit(0); i >= 0; i = aNeighborIdxs.nextSetBit(i + 1)) {
                  final ConceptGraphObject aspect = sortedAspects.get(i);
                  if (add) {
                      //increase the in degree of the aspect since it has one more member within the concept graph
                      aspectInDegrees[i]++;
                  } else {
                      aspect.removeCGNeighbor(member);
                  }
              }
              if (add) {
                  //add the member to the memberScores
                  memberScores.add(member);
                  //update the concept graph score
                  conceptGraphScore += memberScore;
              }
          } //end of the member loop
            if (conceptGraphScore > maxScore) {
                LOGGER.info("\n\n**************" + getId() + ": Keep subgraph score: "  + conceptGraphScore
                    + " previous score " + maxScore + "**************\n");
                //this subgraph is better than all we had before
                aspects.clear();
                for (ConceptGraphObject val : sortedAspects) {
                    aspects.add(val.clone());
                }
                members.clear();
                for (ConceptGraphObject val : memberScores.getObjects()) {
                    members.add(val.clone());
                }
                maxScore = conceptGraphScore;
            }
            //use the basic aspect score to remove the worst aspects based on ALL members first
            final ConceptGraphObject aspect2remove = sortedAspects.remove(sortedAspects.size() - 1);
            //removes also members if their degree is less than the min node degree
            removeNeighbor(aspectNeighbors, null, memberMap, aspect2remove);
            for (ConceptGraphObject aspect : sortedAspects) {
                //reset the remaining aspects for the next round
                aspect.reset();
            }
        }

        //now go backward and collect all members that are connected to the discovered aspects
        //and that improve the graph quality
        final ConceptGraph cg = createConceptGraph(exec, aspects, members, maxScore);
        final List<ConceptGraph> cgs = new ArrayList<ConceptGraph>(1);
        cgs.add(cg);
        exec.setProgress(1.0, "finished");
        return cgs;
    }
}
