/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 07.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.AspectComparator;
import edu.cmu.koetter.util.NeighborBiGraphExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;

/**
 *
 * @author Tobias Koetter
 */
public class FastAspectBasedConceptGraphExtractor extends FastConceptGraphExtractor {

    /**
     * @param view the {@link KPartiteGraphView}
     * @param objectScorer the {@link Scorer}
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the start node
     */
    public FastAspectBasedConceptGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                            final Scorer objectScorer, final int minNodeDegree,
                            final Map<String, NodeNeighbors> neighborMap, final PersistentObject startNode) {
        super(view, objectScorer, minNodeDegree, neighborMap, startNode);
    }

    /**
     * @param exec {@link ExecutionMonitor} to provide progress
     * @return the extracted {@link ConceptGraph}
     * @throws PersistenceException if an exception occurs
     * @throws CanceledExecutionException if the operation has been canceled
     */
    @Override
    protected Collection<ConceptGraph> findConceptGraph(final ExecutionMonitor exec) throws PersistenceException,
            CanceledExecutionException {
        exec.setMessage("analyzing aspects and members...");
        final NeighborBiGraphExtractor extractor = getExtractor();
        final Map<String, Set<String>> aspectNeighbors = extractor.getAspectNeighbors();
        final Map<String, Set<String>> memberNeighbors = extractor.getMemberNeighbors();
        final Map<String, ConceptGraphObject> memberMap =
                createObjectMap(memberNeighbors, aspectNeighbors.keySet(), false);
        final Map<String, ConceptGraphObject> aspectMap =
                createObjectMap(aspectNeighbors, memberNeighbors.keySet(), true);
        //sort the aspects descending based on the density and cue validity with respect to all possible members
        final List<ConceptGraphObject> potentialAspectsSorted = new ArrayList<ConceptGraphObject>(aspectMap.values());
        Collections.sort(potentialAspectsSorted, AspectComparator.getInstance());
        double maxScore = 0;
        final List<ConceptGraph> cgs = new ArrayList<ConceptGraph>(1);
        final Collection<ConceptGraphObject> members = new ArrayList<ConceptGraphObject>();
        final Collection<ConceptGraphObject> aspects = new ArrayList<ConceptGraphObject>();
        while (!potentialAspectsSorted.isEmpty() && !memberMap.isEmpty()) {
            exec.checkCanceled();
            final ScoreCache memberScores = new ScoreCache(memberMap.size(), null);
            for (ConceptGraphObject member : memberMap.values()) {
                final double score = getScorer().computeObjectScore(member, potentialAspectsSorted.size());
                member.setScore(score);
                if (getScorer().improvesScore(memberScores, score)) {
                    memberScores.add(member);
                } else {
                    //remove all members that do not improve the score from the memberScore and the current aspects
                    final Set<String> neighbors = memberNeighbors.get(member.getId());
                    for (final String neighbor : neighbors) {
                        final ConceptGraphObject aspect = aspectMap.get(neighbor);
                        if (aspect != null) {
                            aspect.removeCGNeighbor(member);
                        }
                    }
                }
            }
            //recompute the score for the current aspects based on the optimal member set
            final ScoreCache aspectScores = new ScoreCache(potentialAspectsSorted.size(), null);
            computeScore(potentialAspectsSorted, memberScores.size(), aspectScores);
            //compute the graph score for the current aspect set and its optimal member set
            double score = 0;
            for (ConceptGraphObject member : memberScores) {
                final Set<String> neighbors = memberNeighbors.get(member.getId());
                for (ConceptGraphObject aspect : aspectScores) {
                    final int e;
                    if (neighbors != null && neighbors.contains(aspect.getId())) {
                        e = 1;
                    } else {
                        e = -1;
                    }
                    score += e * aspect.getCueValidity();
                }
            }
            if (score > maxScore) {
//                LOGGER.info("\n\n**************" + getId() + ": Keep subgraph score: "  + score + " previous score "
//                        + maxScore + "**************\n");
                //this subgraph is better than all we had before
                aspects.clear();
                for (ConceptGraphObject val : aspectScores.getObjects()) {
                    aspects.add(val.clone());
                }
                members.clear();
                for (ConceptGraphObject val : memberScores.getObjects()) {
                    members.add(val.clone());
                }
                maxScore = score;
            }
          //use the basic aspect score to remove the worst aspects based on ALL members first
            final ConceptGraphObject aspect2remove = potentialAspectsSorted.remove(potentialAspectsSorted.size() - 1);
            //removes also members if their degree is less than the min node degree
            removeNeighbor(aspectNeighbors, null, memberMap, aspect2remove);
            for (ConceptGraphObject aspect : potentialAspectsSorted) {
                //reset the remaining aspects for the next round
                aspect.reset();
            }
        }

        final ConceptGraph cg = createConceptGraph(exec, aspects, members, maxScore);
        cgs.add(cg);
        exec.setProgress(1.0, "finished");
        return cgs;
    }

//    private List<ConceptGraphObject> createPotentialAspects(final Map<String, ConceptGraphObject> aspectMap,
//        final PersistentObject... startNodes) {
//        final List<ConceptGraphObject> potentialAspects = new ArrayList<ConceptGraphObject>(aspectMap.size());
//        if (aspectMap.isEmpty()) {
//            return potentialAspects;
//        }
//        int minNodeDegree = Integer.MAX_VALUE;
//        for (PersistentObject startNode : startNodes) {
//            final int degree = aspectMap.get(startNode.getId()).getDegree();
//            if (degree < minNodeDegree) {
//                minNodeDegree = degree;
//            }
//        }
//        for (ConceptGraphObject potentialAspect : aspectMap.values()) {
//            if (potentialAspect.getDegree() < minNodeDegree) {
//                //skip all aspects that are more specific than the most specific start node
//                continue;
//            }
//            potentialAspects.add(potentialAspect);
//        }
//        int diff = aspectMap.size() - potentialAspects.size();
//        if (diff > 0) {
//            LOGGER.info(getId() + ": " + diff + " to specific aspects removed of " + aspectMap.size()
//                + " " + (diff * 100) / aspectMap.size() + "(%)");
//        }
//
//        return potentialAspects;
//    }
}
