/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 07.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.NeighborBiGraphExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.NodeLogger;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;

/**
 *
 * @author Tobias Koetter
 */
public class FastMemberBasedConceptGraphExtractor extends FastAspectBasedConceptGraphExtractor {

    private static final NodeLogger LOGGER = NodeLogger.getLogger(FastAspectBasedConceptGraphExtractor.class);

    /**
     * @param view the {@link KPartiteGraphView}
     * @param objectScorer the {@link Scorer}
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the start node
     */
    public FastMemberBasedConceptGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                            final Scorer objectScorer, final int minNodeDegree,
                            final Map<String, NodeNeighbors> neighborMap, final PersistentObject startNode) {
        super(view, objectScorer, minNodeDegree, neighborMap, startNode);
    }


    @Override
    protected Collection<ConceptGraph> findConceptGraph(final ExecutionMonitor exec) throws PersistenceException,
            CanceledExecutionException {
        final NeighborBiGraphExtractor extractor = getExtractor();
        final Map<String, Set<String>> aspectNeighbors = extractor.getAspectNeighbors();
        final Map<String, Set<String>> memberNeighbors = extractor.getMemberNeighbors();
        final Map<String, ConceptGraphObject> memberMap =
                createObjectMap(memberNeighbors, aspectNeighbors.keySet(), false);
        final Map<String, ConceptGraphObject> aspectMap =
                createObjectMap(aspectNeighbors, memberNeighbors.keySet(), true);
        //compute the score for all aspects based on all members
        final ScoreCache basicScores = new ScoreCache(memberMap.size(), new Comparator<ConceptGraphObject>() {
            @Override
            public int compare(final ConceptGraphObject o1, final ConceptGraphObject o2) {
                final int result =
                        Double.compare(o1.getDensity(aspectNeighbors.size()), o2.getDensity(aspectNeighbors.size()));
                if (result == 0) {
                    return o1.getId().compareTo(o2.getId());
                }
                return result;
            }
        });
        computeScore(memberMap.values(), aspectMap.size(), basicScores);
        //force the sorting of the objects in the score cache since they will be altered which changes the sorting order
        basicScores.sort();
        double maxScore = 0;
        final List<ConceptGraph> cgs = new ArrayList<ConceptGraph>(1);
        final Collection<ConceptGraphObject> members = new ArrayList<ConceptGraphObject>();
        final Collection<ConceptGraphObject> aspects = new ArrayList<ConceptGraphObject>();
        while (!aspectMap.isEmpty() && !memberMap.isEmpty()) {
            exec.checkCanceled();
            final ScoreCache aspectScores = new ScoreCache(aspectMap.size(), null);
            for (ConceptGraphObject aspect : aspectMap.values()) {
                final double score = getScorer().computeObjectScore(aspect, memberMap.size());
                aspect.setScore(score);
                if (getScorer().improvesScore(aspectScores, score)) {
                    aspectScores.add(aspect);
                } else {
                    //remove all aspects that do not improve the score
                    final Set<String> neighbors = aspectNeighbors.get(aspect.getId());
                    for (final String neighbor : neighbors) {
                        final ConceptGraphObject member = memberMap.get(neighbor);
                        if (member != null) {
                            member.removeCGNeighbor(aspect);
                        }
                    }
                }
            }
            final ScoreCache memberScores = new ScoreCache(memberMap.size(), null);
            computeScore(memberMap.values(), aspectScores.size(), memberScores);
            //compute the graph score for the current member set and its optimal aspect set
            double score = 0;
            for (ConceptGraphObject aspect : aspectScores) {
                final Set<String> neighbors = aspectNeighbors.get(aspect.getId());
                final double cueValidity = aspect.getCueValidity();
                for (ConceptGraphObject member : memberScores) {
                    final int e;
                    if (neighbors != null && neighbors.contains(member.getId())) {
                        e = 1;
                    } else {
                        e = -1;
                    }
                    score += e * cueValidity * member.getCueValidity();
                }
            }
            if (score > maxScore) {
                LOGGER.info("\n\n**************" + getId() + ": Keep subgraph score: "  + score + " previous score "
                        + maxScore + "**************\n");
                aspects.clear();
                for (ConceptGraphObject val : aspectScores.getObjects()) {
                    aspects.add(val.clone());
                }
                members.clear();
                for (ConceptGraphObject val : memberScores.getObjects()) {
                    members.add(val.clone());
                }
                maxScore = score;
            }
            //use the basic aspect score to remove the worst aspects based on ALL members first
            final ConceptGraphObject member2remove = basicScores.removeObject();
            //removes also members if their degree is less than the min node degree
            removeNeighbor(memberNeighbors, memberMap, aspectMap, member2remove);
            for (ConceptGraphObject member : memberMap.values()) {
                //reset the remaining members for the next round
                member.reset();
            }
        }

        //now go backward and collect all members that are connected to the discovered aspects
        //and that improve the graph quality
        final ConceptGraph cg = createConceptGraph(exec, aspects, members, maxScore);
        cgs.add(cg);
        return cgs;
    }
}
