package edu.cmu.koetter.linkprediction;
/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 03.07.2013 by koetter
 */


import edu.cmu.koetter.conceptgraph.ConceptGraphObject;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 *
 * @author koetter
 */
public class MissingLinks {

    /**
     * Type of missing link.
     * @author koetter
     */
    public enum Type {
        CONCEPT,
        ASPECT;
    }

    /**
     * Class that bundles missing link information.
     * @author Tobias Koetter
     */
    final class MissingLink {

        private final Type m_type;
        private final String m_aspectId;
        private final String m_memberId;
        private final double m_memberOverlap;
        private final double m_aspectCueValidity;
        private final double m_aspectOverlap;
        private final long m_aspectDegree;

        private MissingLink(final Type type, final String aspectId, final String memberId, final double memberOverlap,
            final double aspectCueValidity, final double aspectOverlap, final long aspectDegree) {
            m_type = type;
            m_aspectId = aspectId;
            m_memberId = memberId;
            m_memberOverlap = memberOverlap;
            m_aspectCueValidity = aspectCueValidity;
            m_aspectOverlap = aspectOverlap;
            m_aspectDegree = aspectDegree;
        }

        /**
         * @return the {@link Type} of the missing link
         */
        public Type getType() {
            return m_type;
        }

        /**
         * @return the aspect id
         */
        public String getAspectId() {
            return m_aspectId;
        }

        /**
         * @return the member id
         */
        public String getMemberId() {
            return m_memberId;
        }

        /**
         * @return the member overlap
         */
        public double getMemberOverlap() {
            return m_memberOverlap;
        }

        /**
         * @return the aspectCueValidity
         */
        public double getAspectCueValidity() {
            return m_aspectCueValidity;
        }

        /**
         * @return the aspect overlap
         */
        public double getAspectOverlap() {
            return m_aspectOverlap;
        }

        /**
         * @return the confidence
         */
        public double getConfidence() {
            return m_aspectOverlap * m_memberOverlap;
        }

        /**
         * @return the number of possible members of the aspect
         */
        public long getAspectDegree() {
            return m_aspectDegree;
        }
    }

    private final String[] m_conceptIds;
    private final Map<String, Collection<ConceptGraphObject>> m_missingAspectsPerMember =
            new HashMap<String, Collection<ConceptGraphObject>>();
    private final Map<String, Double> m_potentialMembers = new HashMap<String, Double>();
    private final Collection<String> m_aspectIds;
    private final Collection<String> m_memberIds;
    private final int m_noOfAspects;

    /**
     * @param conceptIds the concept ids
     * @param memberIds the member ids
     * @param aspectIds the aspect ids including the concept ids
     */
    public MissingLinks(final Collection<String> conceptIds, final Collection<String> aspectIds,
        final Collection<String> memberIds) {
        m_conceptIds = conceptIds.toArray(new String[0]);
        m_memberIds = memberIds;
        m_aspectIds = aspectIds;
        m_noOfAspects = m_conceptIds.length + m_aspectIds.size();
    }

    /**
     * @return the concept ids
     */
    public String[] getConceptIds() {
        return m_conceptIds;
    }

    /**
     * @return the size of the concept graph
     */
    public double getSize() {
        return m_noOfAspects * m_memberIds.size();
    }

    /**
     * @return the missing links
     */
    public Collection<MissingLink> getMissingLinks() {
        final Collection<MissingLink> links = new LinkedList<MissingLink>();
        final Set<ConceptGraphObject> aspects = new HashSet<ConceptGraphObject>(m_aspectIds.size());
        double aspectCVSum = 0;
        for (final Entry<String, Collection<ConceptGraphObject>> entry : m_missingAspectsPerMember.entrySet()) {
            final String memberId = entry.getKey();
            final Collection<ConceptGraphObject> missingAspects = entry.getValue();
            //the more aspects the member possesses the higher the score
            final double memberOverlap = (m_noOfAspects - missingAspects.size()) / (double) m_noOfAspects;
            for (final ConceptGraphObject missingAspect : missingAspects) {
                final double cv = missingAspect.getCueValidity();
                //the more members possess the missing aspect the higher is the score
                final double aspectOverlap = missingAspect.getOverlap(m_memberIds.size());
                links.add(new MissingLink(Type.ASPECT, missingAspect.getId(), memberId, memberOverlap, cv,
                    aspectOverlap, missingAspect.getDegree()));
                if (aspects.add(missingAspect)) {
                    aspectCVSum += cv;
                }
            }
        }
        final double avgAspectCv;
        if (aspects.isEmpty()) {
            avgAspectCv = 0;
        } else {
            avgAspectCv = aspectCVSum / aspects.size();
        }
        for (final Entry<String, Double> potentialMember : m_potentialMembers.entrySet()) {
            for (final String concepId : m_conceptIds) {
                links.add(new MissingLink(Type.CONCEPT, concepId, potentialMember.getKey(), potentialMember.getValue(),
                    avgAspectCv, 1, 1));
            }
        }
        return links;
    }

    /**
     * @param aspect the {@link ConceptGraphObject} representation of the aspect that is missing
     * @param memberIds the member ids that aren't linked to the given aspect
     */
    public void addMissingAspectLinks(final ConceptGraphObject aspect, final Collection<String> memberIds) {
        for (final String memberId : memberIds) {
            Collection<ConceptGraphObject> missing = m_missingAspectsPerMember.get(memberId);
            if (missing == null) {
                missing = new LinkedList<ConceptGraphObject>();
                m_missingAspectsPerMember.put(memberId, missing);
            }
            missing.add(aspect);
        }
    }

    /**
     * @param potentialMember the id of the potential member
     * @param confidenceScore the confidence score
     */
    public void addPotentialMember(final String potentialMember, final double confidenceScore) {
        m_potentialMembers.put(potentialMember, Double.valueOf(confidenceScore));
    }
}
