/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 19.04.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.NeighborExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;
import org.knime.core.data.DataCell;
import org.knime.core.data.DataColumnSpec;
import org.knime.core.data.DataColumnSpecCreator;
import org.knime.core.data.DataRow;
import org.knime.core.data.DataTableSpec;
import org.knime.core.data.RowKey;
import org.knime.core.data.collection.CollectionCellFactory;
import org.knime.core.data.collection.ListCell;
import org.knime.core.data.def.BooleanCell;
import org.knime.core.data.def.DefaultRow;
import org.knime.core.data.def.StringCell;
import org.knime.core.node.BufferedDataContainer;
import org.knime.core.node.BufferedDataTable;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionContext;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.InvalidSettingsException;
import org.knime.core.node.KNIMEConstants;
import org.knime.core.node.NodeLogger;
import org.knime.core.node.NodeSettings;
import org.knime.core.node.NodeSettingsRO;
import org.knime.core.node.NodeSettingsWO;
import org.knime.core.node.config.Config;
import org.knime.core.node.config.ConfigRO;
import org.knime.core.node.defaultnodesettings.SettingsModel;
import org.knime.core.node.defaultnodesettings.SettingsModelBoolean;
import org.knime.core.node.defaultnodesettings.SettingsModelInteger;
import org.knime.core.node.defaultnodesettings.SettingsModelIntegerBounded;
import org.knime.core.node.defaultnodesettings.SettingsModelString;
import org.knime.core.util.ThreadPool;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.knime.GraphRepository;
import org.knime.network.core.knime.NetCol;
import org.knime.network.core.knime.cell.GraphCellFactory;
import org.knime.network.core.knime.node.AbstractGraphNodeModel;
import org.knime.network.core.knime.node.KPartiteGraphViewAndTable2TableNodeModel;
import org.knime.network.core.knime.port.GraphPortObjectSpec;

/**
 *
 * @author Tobias Koetter
 */
public class ConceptGraphExtractorNodeModel extends KPartiteGraphViewAndTable2TableNodeModel {

    /**Member ids column name.*/
    public static final String COL_MEMBER_IDS = "Member ids";


    /**Aspect ids column name.*/
    public static final String COL_ASPECT_IDS = "Aspect ids";


    /**Concept ids column name.*/
    public static final String COL_CONCEPT_IDS = "Concept ids";


    private static final NodeLogger LOGGER = NodeLogger.getLogger(ConceptGraphExtractorNodeModel.class);


    private static final String CFG_FILE = "settingsFile.xml";

    private static final String CFG_SETTING = "saveInternalsSettings";

    private static final String CFG_NETWORK_UUIDS = "networkUUIDs";

    private final LinkedList<String> m_netCacheIDs = new LinkedList<String>();

    private final SettingsModelString m_objectIdColumn = createObjectIdColumnModel();
    private final SettingsModelString m_score = createScoreModel();
    private final SettingsModelString m_type = createTypeModel();
    private final SettingsModelInteger m_maxNoOfProcesses = createNoOfProcessesModel();
    private final SettingsModelInteger m_cueValidity = createCueValidityModel();
    private final SettingsModelInteger m_density = createDensityModel();
    private final SettingsModelInteger m_size = createSizeModel();
    private final SettingsModelInteger m_minNodeDegree = createMinNodeDegreeModel();
    private final SettingsModelBoolean m_addGraphCells = createAddGraphCellsModel();
    private final SettingsModelBoolean m_cacheGraphCells = AbstractGraphNodeModel.createCacheGraphCellModel();

    private final SettingsModel [] m_models = new SettingsModel[] {m_objectIdColumn, m_score, m_maxNoOfProcesses,
            m_type, m_cueValidity, m_density, m_size, m_minNodeDegree, m_addGraphCells, m_cacheGraphCells};

    /**
     * @return the maximum number of processes model
     */
    static SettingsModelInteger createNoOfProcessesModel() {
        return new SettingsModelIntegerBounded("maximumNoOfProcesses",
                                           KNIMEConstants.GLOBAL_THREAD_POOL.getMaxThreads(), 1, Integer.MAX_VALUE);
    }

    /**
     * @return the score model
     */
    static SettingsModelString createScoreModel() {
        return new SettingsModelString("score", ConceptGraphScore.SUM.getActionCommand());
    }


    /**
     * @return the type model
     */
    static SettingsModelString createTypeModel() {
        return new SettingsModelString("extractorType", ConceptGraphExtractorType.ITERATIVE.getActionCommand());
    }

    /**
     * @return the minimum node degree model
     */
    static SettingsModelInteger createMinNodeDegreeModel() {
        return new SettingsModelIntegerBounded("minNodeDegree", 1, 1, Integer.MAX_VALUE);
    }

    /**
     * @return size exponent model
     */
    static SettingsModelInteger createSizeModel() {
        return new SettingsModelIntegerBounded("size", 1, 0, Integer.MAX_VALUE);
    }

    /**
     * @return density exponent model
     */
    static SettingsModelInteger createDensityModel() {
        return new SettingsModelIntegerBounded("density", 1, 0, Integer.MAX_VALUE);
    }

    /**
     * @return cue exponent power model
     */
    static SettingsModelInteger createCueValidityModel() {
        return new SettingsModelIntegerBounded("cueValidity", 1, 0, Integer.MAX_VALUE);
    }

    /**
     * @return the add graph cells model
     */
    static SettingsModelBoolean createAddGraphCellsModel() {
        return new SettingsModelBoolean("addGraphCells", true);
    }

    /**
     * @return the object id model
     */
    static SettingsModelString createObjectIdColumnModel() {
        return new SettingsModelString("objectIdColumn", null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected BufferedDataTable execute(final ExecutionContext exec,
                final KPartiteGraphView<PersistentObject, Partition> view, final BufferedDataTable table)
                        throws Exception {
        exec.setMessage("Searching...");
        //use cluster Pruning factor of infinite or number of objects
        final BufferedDataContainer dc = exec.createDataContainer(createSpec());
        if (table.getRowCount() < 1) {
            setWarningMessage("Empty input table or network");
            dc.close();
            return dc.getTable();
        }
        final String colName = m_objectIdColumn.getStringValue();
        final int colIdx = table.getSpec().findColumnIndex(colName);
        if (colIdx < 0) {
            throw new InvalidSettingsException("Column " + colName + " not found in input table");
        }
        final int procCount = m_maxNoOfProcesses.getIntValue();
        final NeighborExtractor extractor = new NeighborExtractor(view);
        //extract the neighbors of all nodes to speedup the node
        final Map<String, NodeNeighbors> neighborMap =
                extractor.getNodeNeighbors(exec.createSubProgress(0.05), procCount);
        //initialize the thread pool
        final ThreadPool pool =
            KNIMEConstants.GLOBAL_THREAD_POOL.createSubPool();
        //The semaphore restricts the number of concurrent processes
        final Semaphore semaphore = new Semaphore(procCount);
        if (procCount == 1) {
            //acquire the only permit to prevent memory problems
            semaphore.acquire();
        }
        final Queue<Future<Collection<ConceptGraph>>> futures = new LinkedList<Future<Collection<ConceptGraph>>>();
        final int noOfRows = table.getRowCount();
        for (final DataRow row : table) {
            final ExecutionContext subExec = exec.createSubExecutionContext(0.95 / noOfRows);
            exec.checkCanceled();
            final DataCell conceptCell = row.getCell(colIdx);
            if (conceptCell.isMissing()) {
                continue;
            }
            final PersistentObject startNode = getNode(view, conceptCell);
            final ConceptGraphScore score = ConceptGraphScore.get(m_score.getStringValue());
            final Scorer objectScorer =
                    score.getScorer(m_density.getIntValue(), m_cueValidity.getIntValue(), m_size.getIntValue());
            final ConceptGraphExtractorType extractorType = ConceptGraphExtractorType.get(m_type.getStringValue());
            final ConceptGraphExtractor cgExtractor = extractorType.getExtractor(
                view, objectScorer, m_minNodeDegree.getIntValue(), neighborMap, startNode);
            futures.add(pool.enqueue(new ConceptGraphExtractorCallable(subExec, semaphore, cgExtractor)));
        }
        if (procCount == 1) {
            //release the only permit to start the Runnables
            semaphore.release();
        }
        int counter = 0;
        int cgCounter = 0;
        while (!futures.isEmpty()) {
            //get AND remove the finished future from the queue
            final Future<Collection<ConceptGraph>> future = futures.poll();
            exec.checkCanceled();
            counter++;
            final Collection<ConceptGraph> cgs = future.get();
            exec.setProgress(counter / (double) noOfRows, counter + " of " + noOfRows + " rows analyzed");
            for (final ConceptGraph cg : cgs) {
                if (cg == null) {
                    continue;
                }
                final List<DataCell> cells = new LinkedList<DataCell>();
                cells.add(new StringCell(cg.getId()));
                cells.add(new StringCell(cg.getStartNode().getId()));
                cells.add(createListCell(cg.getConceptNodes()));
                cells.add(createListCell(cg.getAspectNodes()));
                cells.add(createListCell(cg.getMemberNodes()));
                cells.add(BooleanCell.get(
                    Arrays.equals(new String[] {cg.getStartNode().getId()}, cg.getConceptNodes())));
                cells.addAll(cg.getScoreCells());
                if (m_addGraphCells.getBooleanValue()) {
                    final KPartiteGraphView<PersistentObject, Partition> graph = cg.getConceptGraph(neighborMap);
                    if (m_cacheGraphCells.getBooleanValue()) {
                        GraphRepository.getInstance().cacheGraphCell(graph);
                        m_netCacheIDs.add(graph.getUUID());
                    }
                    cells.add(GraphCellFactory.createCell(graph));
                }
                final DataRow row = new DefaultRow(RowKey.createRowKey(cgCounter++), cells);
                dc.addRowToTable(row);
            }
        }
        dc.close();
        return dc.getTable();
    }

    /**
     * @return
     */
    private DataTableSpec createSpec() {
        final Collection<DataColumnSpec> specs = new LinkedList<DataColumnSpec>();
        final DataColumnSpecCreator creator =
                new DataColumnSpecCreator("Id", StringCell.TYPE);
        specs.add(creator.createSpec());
        creator.setName("Start node id");
        specs.add(creator.createSpec());
        creator.setName(COL_CONCEPT_IDS);
        creator.setType(ListCell.getCollectionType(StringCell.TYPE));
        specs.add(creator.createSpec());
        creator.setName(COL_ASPECT_IDS);
        specs.add(creator.createSpec());
        creator.setName(COL_MEMBER_IDS);
        specs.add(creator.createSpec());
        creator.setName("Start nodes equal concept nodes");
        creator.setType(BooleanCell.TYPE);
        specs.add(creator.createSpec());
        specs.addAll(ConceptGraph.createScoreSpecs());
        if (m_addGraphCells.getBooleanValue()) {
            creator.setName(NetCol.GRAPH);
            creator.setType(GraphCellFactory.getType());
            specs.add(creator.createSpec());
        }
        return new DataTableSpec(specs.toArray(new DataColumnSpec[2]));
    }

    /**
     * @param vals the {@link String} values
     * @return a {@link ListCell} that contains the given {@link String}s
     */
    public static DataCell createListCell(final String... vals) {
        final Collection<StringCell> cells = new ArrayList<StringCell>(vals.length);
        for (final String node : vals) {
            cells.add(new StringCell(node));
        }
        return CollectionCellFactory.createListCell(cells);
    }

    /**
     * @param view
     * @param cell
     * @return
     * @throws PersistenceException
     * @throws InvalidSettingsException
     */
    private PersistentObject getNode(final KPartiteGraphView<PersistentObject, Partition> view, final DataCell cell)
            throws PersistenceException, InvalidSettingsException {
        final String id = cell.toString();
        final PersistentObject node = view.getNode(id);
        if (node == null) {
            throw new InvalidSettingsException("node with id " + id + " not found in network");
        }
        return node;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected DataTableSpec getTableSpec(final GraphPortObjectSpec viewSpec, final DataTableSpec tableSpec)
            throws InvalidSettingsException {
        if (tableSpec.getNumColumns() < 0) {
            throw new InvalidSettingsException("Empty input table");
        }
        final String colName = m_objectIdColumn.getStringValue();
        if (colName == null || colName.isEmpty()) {
            final String presetName = tableSpec.getColumnNames()[0];
            m_objectIdColumn.setStringValue(presetName);
            setWarningMessage("Column preset to " + presetName);
        } else if (!tableSpec.containsName(colName)) {
            throw new InvalidSettingsException("Column with name: " + colName + " not found in input table");
        }
        return createSpec();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void saveSettingsTo(final NodeSettingsWO settings) {
        for (final SettingsModel model : m_models) {
            model.saveSettingsTo(settings);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void validateSettings(final NodeSettingsRO settings) throws InvalidSettingsException {
        final String colName =
                ((SettingsModelString)m_objectIdColumn.createCloneWithValidatedValue(settings)).getStringValue();
        if (colName == null || colName.isEmpty()) {
            throw new InvalidSettingsException("No object id column selected");
        }
//        for (SettingsModel model : m_models) {
//            model.validateSettings(settings);
//        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void loadValidatedSettingsFrom(final NodeSettingsRO settings) throws InvalidSettingsException {
        for (final SettingsModel model : m_models) {
            try {
                model.loadSettingsFrom(settings);
            } catch (final InvalidSettingsException e) {
                // old setting ignore the exception
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void loadInternals(final File nodeInternDir,
            final ExecutionMonitor exec)
    throws IOException, CanceledExecutionException {
       final File settingFile = new File(nodeInternDir, CFG_FILE);
        final FileInputStream inData = new FileInputStream(settingFile);
        final ConfigRO config = NodeSettings.loadFromXML(inData);
        try {
            final String[] networkUUIDs =
                config.getStringArray(CFG_NETWORK_UUIDS);
            if (networkUUIDs.length > 0) {
                m_netCacheIDs.addAll(Arrays.asList(networkUUIDs));
            }
        } catch (final InvalidSettingsException e) {
            throw new IOException("Invalid loadInternals settings file",
                    e.getCause());
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void saveInternals(final File nodeInternDir,
            final ExecutionMonitor exec)
    throws IOException, CanceledExecutionException {
        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("Entering saveInternals(nodeInternDir, exec) "
                    + "of class Network2RowNodeModel.");
        }
        try {
            final File settingFile = new File(nodeInternDir, CFG_FILE);
            final FileOutputStream dataOS = new FileOutputStream(settingFile);
            final Config config = new NodeSettings(CFG_SETTING);
            exec.checkCanceled();
            config.addStringArray(CFG_NETWORK_UUIDS,
                    m_netCacheIDs.toArray(new String[0]));
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug("Exiting saveInternals(nodeInternDir, exec) "
                        + "of class Network2RowNodeModel.");
            }
            config.saveToXML(dataOS);
        } catch (final Exception e) {
            throw new IOException(e.getMessage(), e.getCause());
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void resetInternal() {
        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug("Entering reset() of class Network2RowNodeModel.");
        }
        for (final String id : m_netCacheIDs) {
            //detach all networks that were created by this node from the
            //cache
            GraphRepository.getInstance().detach(id);
        }
        m_netCacheIDs.clear();
        if (LOGGER.isDebugEnabled()) {
            LOGGER.debug(
                    "Exiting resetInternal() of class Network2RowNodeModel.");
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void onDispose() {
        try {
           resetInternal();
        } catch (final Exception e) {
            LOGGER.debug("Exception while finalizing BisonetNode: "
                    + e.getMessage());
        }
    }
}
