/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 25.04.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Set;

/**
 *
 * @author Tobias Koetter
 */
public class ConceptGraphObject {

    /**Ascending order based on the score.*/
    public static final Comparator<ConceptGraphObject> ASC_SCORE_COMP = new Comparator<ConceptGraphObject>() {
        @Override
        public int compare(final ConceptGraphObject o1, final ConceptGraphObject o2) {
            final int result = Double.compare(o1.getScore(), o2.getScore());
            if (result == 0) {
                return o1.getId().compareTo(o2.getId());
            }
            return result;
        }
    };


    /**Ascending order based on the score.*/
    public static final Comparator<ConceptGraphObject> DESC_CV_COMP = new Comparator<ConceptGraphObject>() {
        @Override
        public int compare(final ConceptGraphObject o1, final ConceptGraphObject o2) {
            final int result = Double.compare(o2.getCueValidity(), o1.getCueValidity());
            if (result == 0) {
                return o2.getId().compareTo(o1.getId());
            }
            return result;
        }
    };

    private final String m_id;
    /**Number of neighbors within the concept graph.*/
    private final int m_initialInDegree;
    /**Number of neighbors outside of the concept graph.*/
    private final int m_initialOutDegree;
    private final boolean m_aspect;
    /**The number of neighbors that have been removed/added from the concept graph after the initialization.
     * If the value is positive more neighbors were removed than added. If it is negative more neighbors have
     * been added than removed.*/
    private int m_removedCGNeighbors = 0;
    private double m_score = 0;

    /**
     * @param id id
     * @param inDegree in degree
     * @param outDegree out degree
     * @param aspect <code>true</code> if the node is an aspect
     */
    public ConceptGraphObject(final String id, final int inDegree, final int outDegree, final boolean aspect) {
        this(id, inDegree, outDegree, 0, 0, aspect);
    }

    private ConceptGraphObject(final String id, final int inDegree, final int outDegree, final int removedGCNeighbors,
        final double score, final boolean aspect) {
        if (id == null) {
            throw new IllegalArgumentException("id must not be null");
        }
        m_id = id;
        m_initialInDegree = inDegree;
        m_initialOutDegree = outDegree;
        m_removedCGNeighbors = removedGCNeighbors;
        m_score = score;
        m_aspect = aspect;
    }

    /**
     * @return the nodeId
     */
    public String getId() {
        return m_id;
    }

    /**
     * @return the number of neighbors within the concept graph
     */
    public int getInDegree() {
        return m_initialInDegree - m_removedCGNeighbors;
    }

    /**
     * @return the number of neighbors outside of the concept graph
     */
    public int getOutDegree() {
        return m_initialOutDegree + m_removedCGNeighbors;
    }


    /**
     * @return the total number of neighbors
     */
    public int getDegree() {
        return m_initialInDegree + m_initialOutDegree;
    }

    /**
     * @return the aspect
     */
    public boolean isAspect() {
        return m_aspect;
    }

    /**
     * @return the cue validity
     */
    public double getCueValidity() {
        final double cueValidity = getInDegree() / (double) getDegree();
        return cueValidity;
    }

    /**
     * @return the stronger cue validity
     */
    public double getStrongCueValidity() {
        final double cueValidity = (getInDegree() - getOutDegree()) / (double) getDegree();
        return cueValidity;
    }

    /**
     * @param size the size of the neighbor set
     * @return the overlap = number of set / size
     */
    public double getOverlap(final long size) {
        if (size == 0) {
            return 0;
        }
        final double overlap = getInDegree() / (double) size;
        return overlap;
    }


    /**
     * @param size the size of the neighbor set
     * @return the density = (number of set - number of unset) / size
     */
    public double getDensity(final long size) {
        if (size == 0) {
            return 0;
        }
//        final double density = (m_inDegree - (size - m_inDegree)) / (double) size; same as
        final double density = -1 + 2 * getInDegree() / (double) size;
        return density;
    }

    /**
     * @param score the score of this object
     */
    public void setScore(final double score) {
        m_score  = score;
    }

    /**
     * @return the score
     */
    public double getScore() {
        return m_score;
    }

    /**
     * @param object2Remove the {@link ConceptGraphObject} to remove
     */
    public void removeCGNeighbor(final ConceptGraphObject object2Remove) {
        m_removedCGNeighbors++;
    }

    /**
     * @param object2Add the {@link ConceptGraphObject} to add
     */
    public void addCGNeighbor(final ConceptGraphObject object2Add) {
        m_removedCGNeighbors--;
    }

    /**
     * Resets the removed neighbors and the score.
     */
    public void reset() {
        m_removedCGNeighbors = 0;
        m_score = 0;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public ConceptGraphObject clone() {
        return new ConceptGraphObject(m_id, m_initialInDegree, m_initialOutDegree, m_removedCGNeighbors,
            m_score, m_aspect);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();
        builder.append(m_id);
        builder.append(", ");
        builder.append("score=");
        builder.append(m_score);
        builder.append(", ");
        builder.append("inD=");
        builder.append(getInDegree());
        builder.append(", outD=");
        builder.append(getOutDegree());
        builder.append(", ");
        if (m_aspect) {
            builder.append("aspect");
        } else {
            builder.append("member");
        }
        return builder.toString();
    }

    /**
     * @param objects {@link Collection} of {@link ConceptGraphObject} to create the id set for
     * @return {@link Set} with the ids of all the {@link ConceptGraphObject}s
     */
    public static Set<String> createIdSet(final Collection<ConceptGraphObject> objects) {
        final Set<String> set = new HashSet<String>(objects.size());
        for (ConceptGraphObject object : objects) {
            set.add(object.getId());
        }
        return set;
    }

    /**
     * @param objects the {@link ConceptGraphObject} to extract the ids
     * @param sort <code>true</code> if it should be sorted by id
     * @return the ids of the {@link ConceptGraphObject}s
     */
    public static String[] createIdArray(final Collection<ConceptGraphObject> objects, final boolean sort) {
        final String[] ids = new String[objects.size()];
        int idx = 0;
        for (ConceptGraphObject object : objects) {
            ids[idx++] = object.getId();
        }
        if (sort) {
            Arrays.sort(ids);
        }
        return ids;
    }
}
