/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 14.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.NeighborBiGraphExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.NodeLogger;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.util.GraphUtil;

/**
 *
 * @author koetter
 */
public abstract class FastConceptGraphExtractor implements ConceptGraphExtractor {

    private static final NodeLogger LOGGER = NodeLogger.getLogger(FastAspectBasedConceptGraphExtractor.class);
    private final KPartiteGraphView<PersistentObject, Partition> m_view;
    private final NeighborBiGraphExtractor m_extractor;
    private final PersistentObject m_startNode;
    private String m_id;
    private final Scorer m_scorer;
    private final int m_minNodeDegree;
    private final Map<String, NodeNeighbors> m_neighborMap;

    /**
     * @param view the {@link KPartiteGraphView}
     * @param objectScorer the {@link Scorer}
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the start node
     */
    public FastConceptGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                                     final Scorer objectScorer, final int minNodeDegree,
                                     final Map<String, NodeNeighbors> neighborMap, final PersistentObject startNode) {
        m_view = view;
        m_scorer = objectScorer;
        m_neighborMap = neighborMap;
        if (minNodeDegree < 1) {
            throw new IllegalArgumentException("Minimum node degree must be larger than 0");
        }
        m_minNodeDegree = minNodeDegree;
        m_startNode = startNode;
        m_id = startNode.getId();
        m_extractor = new NeighborBiGraphExtractor(view, neighborMap);
    }

    /**
     * @return the {@link NeighborBiGraphExtractor} to use
     */
    protected NeighborBiGraphExtractor getExtractor() {
        return m_extractor;
    }

    /**
     * @return the view
     */
    public KPartiteGraphView<PersistentObject, Partition> getView() {
        return m_view;
    }

    /**
     * @return {@link Map} with node {@link NodeNeighbors}s
     */
    protected Map<String, NodeNeighbors> getNeighborMap() {
        return m_neighborMap;
    }

    /**
     * @return the startNodes
     */
    public PersistentObject getStartNode() {
        return m_startNode;
    }

    /**
     * @return the id
     */
    public String getId() {
        return m_id;
    }

    /**
     * @return the {@link Scorer} to use
     */
    protected Scorer getScorer() {
        return m_scorer;
    }

    /**
     * @return the minimum node degree
     */
    protected int getMinNodeDegree() {
        return m_minNodeDegree;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Collection<ConceptGraph> extractConceptGraph(final ExecutionMonitor exec)
            throws PersistenceException, CanceledExecutionException {
        final String id = getId();
        exec.setMessage("Extracting neighborhood of " + id);
        final long startTime = System.currentTimeMillis();
        getExtractor().run(exec.createSubProgress(0.6), getMinNodeDegree(), getStartNode());
        final long bigraphTime = System.currentTimeMillis() - startTime;
        exec.checkCanceled();
        exec.setMessage("Creating concept graph for " + id);
        final long startTime2 = System.currentTimeMillis();
        final Collection<ConceptGraph> cgs = findConceptGraph(exec.createSubProgress(0.4));
        final long cgTime = System.currentTimeMillis() - startTime2;
        exec.checkCanceled();
        exec.setProgress(1.0, "Concept graph created for " + id);
        final long totalTime = bigraphTime + cgTime;
        final long milliSeconds = totalTime % 1000;
        long seconds = Math.round(totalTime / 1000.0);
        long minutes = seconds / 60;
        seconds %= 60;
        long hours = minutes / 60;
        minutes %= 60;
        long days = hours / 24;
        hours %= 24;
        StringBuilder msg = new StringBuilder();
        msg.append("\t*****" + id);
        msg.append(" #a: " + getExtractor().getAspectNeighbors().size());
        msg.append(" #m: " + getExtractor().getMemberNeighbors().size());
        msg.append(" Extraction: " + bigraphTime + " ms " + (int)(bigraphTime / (double) totalTime * 100) + "%");
        msg.append(" Discovery: " + cgTime + " ms " + (int)(cgTime / (double) totalTime * 100) + "%");
        msg.append(" Total: ");
        if (days != 0) { msg.append(days); msg.append(" d "); }
        if (hours != 0) { msg.append(hours); msg.append(" h "); }
        if (minutes != 0) { msg.append(minutes); msg.append(" m "); }
        if (seconds != 0) { msg.append(seconds); msg.append(" s "); }
        msg.append(milliSeconds); msg.append(" ms");
        LOGGER.debug(msg);
        exec.setProgress(1.0, "Concept graph created for " + id + " in " + totalTime + " ms");
        return cgs;
    }

    /**
     * @param exec {@link ExecutionMonitor} to provide progress
     * @return the extracted {@link ConceptGraph}
     * @throws PersistenceException if an exception occurs
     * @throws CanceledExecutionException if the operation has been canceled
     */
    protected abstract Collection<ConceptGraph> findConceptGraph(ExecutionMonitor exec)
            throws PersistenceException, CanceledExecutionException;

    /**
     * @param objects the {@link ConceptGraphObject}s to compute the score for
     * @param neighborSetSize the size of the corresponding neighbor set
     * @param scores the {@link ScoreCache} to use
     */
    protected void computeScore(final Collection<ConceptGraphObject> objects, final int neighborSetSize,
                                final ScoreCache scores) {
        scores.reset();
        for (ConceptGraphObject object : objects) {
            object.setScore(getScorer().computeObjectScore(object, neighborSetSize));
            scores.add(object);
        }
    }

    /**
     * @param object2RemoveNeighborMap the {@link Map} that contains the neighbors of the object to remove
     * @param map2Remove the {@link Map} to remove the object from or <code>null</code> if it shouldn't be removed
     * @param neighborMap2Update the {@link Map} that contains the neighbor to update
     * @param object2Remove the {@link ConceptGraphObject} to remove
     */
    protected void removeNeighbor(final Map<String, Set<String>> object2RemoveNeighborMap,
                                  final Map<String, ConceptGraphObject> map2Remove,
                                  final Map<String, ConceptGraphObject> neighborMap2Update,
                                  final ConceptGraphObject object2Remove) {
        if (map2Remove != null) {
            map2Remove.remove(object2Remove.getId());
        }
        final Set<String> neighbors = object2RemoveNeighborMap.get(object2Remove.getId());
        for (String neighbor : neighbors) {
            final ConceptGraphObject object = neighborMap2Update.get(neighbor);
            if (object != null) {
                object.removeCGNeighbor(object2Remove);
            }
        }
    }

    /**
     * @param candidates the candidates and their neighbors
     * @param base the base set of neighbors to compare with
     * @param isAspect <code>true</code> if the candidates are the aspects <code>false</code> if they are the members
     * @return a {@link Map} with the given candidates as key and their corresponding {@link ConceptGraphObject}
     * as value
     */
    public static Map<String, ConceptGraphObject> createObjectMap(final Map<String, Set<String>> candidates,
                                                              final Set<String> base, final boolean isAspect) {
        final HashMap<String, ConceptGraphObject> map =
                new HashMap<String, ConceptGraphObject>(candidates.size());
        for (Entry<String, Set<String>> candidate : candidates.entrySet()) {
            final Set<String> neighbors = candidate.getValue();
            final int inDegree = GraphUtil.computeOverlap(base, neighbors);
            final int outDegree = neighbors.size() - inDegree;
            final ConceptGraphObject object = new ConceptGraphObject(candidate.getKey(), inDegree, outDegree, isAspect);
            map.put(candidate.getKey(), object);
        }
        return map;
    }

    /**
     * This method should be called to create the {@link ConceptGraph} at the end of the
     * {@link #findConceptGraph(ExecutionMonitor)} method.
     * @param exec {@link ExecutionMonitor} to provide progress
     * @param aspects the aspects
     * @param members the members
     * @param score the score of the concept graph
     * @return the {@link ConceptGraph}
     * @throws CanceledExecutionException if the operation has been canceled
     */
    protected ConceptGraph createConceptGraph(final ExecutionMonitor exec,
        final Collection<ConceptGraphObject> aspects, final Collection<ConceptGraphObject> members, final double score)
                throws CanceledExecutionException {
        //now go backward and collect all members that are connected to the discovered aspects
        //and that improve the graph quality
        final NeighborBiGraphExtractor extractor = getExtractor();
        final int noOfPossibleMembers = extractor.getMemberNeighbors().size();
        final int noOfPossibleAspects = extractor.getAspectNeighbors().size();
        final ConceptGraph cg = new ConceptGraph(getView(), aspects, members, score, getStartNode(),
            noOfPossibleAspects, noOfPossibleMembers);
        return cg;
    }
}