/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 21.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.AspectComparator;
import edu.cmu.koetter.util.MemberComparator;
import edu.cmu.koetter.util.NeighborBiGraphExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.NodeLogger;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.util.GraphUtil;

/**
 *
 * @author Tobias Koetter
 */
public class IterativeConceptGraphExtractor extends FastConceptGraphExtractor {

    private static final NodeLogger LOGGER = NodeLogger.getLogger(IterativeConceptGraphExtractor.class);

    private static final int MAX_ITERATION = 1000;

    private Map<String, Set<String>> m_aspectNeighbors;
    private Map<String, Set<String>> m_memberNeighbors;
    private boolean m_changed = false;
    private double m_score = 0;
    private Map<String, ConceptGraphObject> m_members;
    private ArrayList<ConceptGraphObject> m_aspects;
    private List<ConceptGraphObject> m_potentialAspectsSorted;
    private List<ConceptGraphObject> m_potentialMembersSorted;

    private boolean m_useCueValidity;


    /**
     * @param view the {@link KPartiteGraphView}
     * @param scorer the {@link Scorer}
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the start node
     */
    public IterativeConceptGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
        final Scorer scorer, final int minNodeDegree, final Map<String, NodeNeighbors> neighborMap,
        final PersistentObject startNode) {
        super(view, scorer, minNodeDegree, neighborMap, startNode);
        m_useCueValidity = scorer.useCueValidity();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected Collection<ConceptGraph> findConceptGraph(final ExecutionMonitor exec) throws PersistenceException,
        CanceledExecutionException {
        exec.setMessage("analyzing aspects and members...");
        final long startTime = System.currentTimeMillis();
        final NeighborBiGraphExtractor extractor = getExtractor();
        m_aspectNeighbors = extractor.getAspectNeighbors();
        m_memberNeighbors = extractor.getMemberNeighbors();
        m_members = createObjectMap(m_memberNeighbors, m_aspectNeighbors.keySet(), false);
        m_aspects = new ArrayList<ConceptGraphObject>(
                createObjectMap(m_aspectNeighbors, m_memberNeighbors.keySet(), true).values());
//        final Set<String> aspectIds = m_aspectNeighbors.keySet();
//        final Set<String> memberIds = m_memberNeighbors.keySet();
        m_potentialMembersSorted = new LinkedList<ConceptGraphObject>();
//                createList(m_memberNeighbors, memberIds, aspectIds, MemberComparator.getInstance(), false);
        m_potentialAspectsSorted = new LinkedList<ConceptGraphObject>();
//                createList(m_aspectNeighbors, aspectIds, memberIds, AspectComparator.getInstance(), true);
        m_score = computeInitialScore();
        m_changed = true;
        int iterationCounter = 0;
        while (m_changed && iterationCounter < MAX_ITERATION) {
            exec.setMessage("improvement iteration: " + ++iterationCounter);
            exec.checkCanceled();
            m_changed = false;
            if (!m_potentialAspectsSorted.isEmpty()) {
                addAspects();
            }
            if (!m_potentialMembersSorted.isEmpty()) {
                addMembers();
            }
            if (removeAspects()) {
                Collections.sort(m_potentialAspectsSorted, AspectComparator.getInstance());
            }
            if (removeMembers()) {
                Collections.sort(m_potentialMembersSorted, MemberComparator.getInstance());
            }
        }
        final long endTime = System.currentTimeMillis();
        final long durationTime = endTime - startTime;
        LOGGER.info(getStartNode().getId() + ": Time for improvement over " + iterationCounter
            + " iterations: " + durationTime + " ms");
        final LinkedList<ConceptGraph> cgs = new LinkedList<ConceptGraph>();
        final int noOfPossibleMembers = m_memberNeighbors.size();
        final int noOfPossibleAspects = m_aspectNeighbors.size();
        cgs.add(new ConceptGraph(getView(), m_aspects, m_members.values(), m_score, getStartNode(),
            noOfPossibleAspects, noOfPossibleMembers));
        return cgs;
    }


    /**
     * @return
     */
    private double computeInitialScore() {
      //recompute the score for the current aspects based on the optimal member set
        final ScoreCache aspectScores = new ScoreCache(m_aspectNeighbors.size(), null);
        computeScore(m_aspects, m_memberNeighbors.size(), aspectScores);
        //compute the graph score for the current aspect set and its optimal member set
        double score = 0;
        for (ConceptGraphObject member : m_members.values()) {
            final Set<String> neighbors = m_memberNeighbors.get(member.getId());
            for (ConceptGraphObject aspect : aspectScores) {
                final int e;
                if (neighbors != null && neighbors.contains(aspect.getId())) {
                    e = 1;
                } else {
                    e = -1;
                }
                score += e * aspect.getCueValidity();
            }
        }
        return score;
    }

    /**
     * @return
     */
    private boolean removeMembers() {
        boolean changed = false;
        final BitSet aNeighborIdxs = new BitSet(m_aspects.size());
        final List<ConceptGraphObject> members = new ArrayList<ConceptGraphObject>(m_members.values());
        //we have to check the length of the sorted members in each round since it might have changed
        for (int mIdx = 0; mIdx < members.size(); mIdx++) {
            final ConceptGraphObject member = members.get(mIdx);
            double cueValDiff = 0;
            double mScore = 0;
            aNeighborIdxs.clear();
            final Set<String> mNeighbors = m_memberNeighbors.get(member.getId());
            for (int idx = 0, length = m_aspects.size(); idx < length; idx++) {
                final ConceptGraphObject aspect = m_aspects.get(idx);
                final double aInDegree = aspect.getInDegree();
                final double aDegree = aspect.getDegree();
                if (mNeighbors.contains(aspect.getId())) {
                    //the member is a neighbor of the aspect
                    mScore += (aInDegree + 1) / aDegree;
                    if (m_useCueValidity) {
                        //this is the cue validity diff of all members that possess the aspect minus the members
                        //that do not possess the aspect
                        //the cue validity affects only the aspects that are neighbors of the new member
                        cueValDiff += (aInDegree - (m_members.size() - aInDegree)) / aDegree;
                    }
                    aNeighborIdxs.set(idx);
                } else {
                    //the aspect is no member of the neighbor so we do not need to compute the cue validity diff
                    mScore -= aInDegree / aDegree;
                }
            }
            final double memberScore = mScore + cueValDiff;
            if (memberScore < 0) {
                changed = true;
                //loop over all aspects that are affected by the current member
                for (int idx = aNeighborIdxs.nextSetBit(0); idx >= 0; idx = aNeighborIdxs.nextSetBit(idx + 1)) {
                    final ConceptGraphObject aspect = m_aspects.get(idx);
                    aspect.removeCGNeighbor(member);
                }
                //since we remove the member from the sorted member list we have to decrease also the counter by one
                removeMember(member);
                members.remove(mIdx--);
                //update the concept graph score
                m_score -= memberScore;
            }
        }
        return changed;
    }

    /**
     * @return
     */
    private boolean removeAspects() {
      //we have to check the length of the sorted aspects in each round since it might have changed
        boolean changed = false;
        for (int aIdx = 0; aIdx < m_aspects.size(); aIdx++) {
            final ConceptGraphObject aspect = m_aspects.get(aIdx);
            final int aInDegree = aspect.getInDegree();
            final int existingMinusNotExisting = aInDegree - (m_members.size() - aInDegree);
            if (existingMinusNotExisting < 0) {
                final Set<String> aNeighbors = m_aspectNeighbors.get(aspect.getId());
                //this aspect decreases the score since it is an aspect with less than half of the members
                for (String memberId : aNeighbors) {
                    ConceptGraphObject member = m_members.get(memberId);
                    if (member != null) {
                        //add the aspect to its neighbors
                        member.removeCGNeighbor(aspect);
                    }
                }
                //since we remove the potential aspect from the sorted potential aspect list we have to
                //decrease also the counter by one
                removeAspect(aIdx--, aspect);
                //update the concept graph score
                m_score -= existingMinusNotExisting * aInDegree / (double) aspect.getDegree();
                changed = true;
            }
        }
        return changed;
    }

    private void addAspects() {
        //we have to check the length of the sorted aspects in each round since it might have changed
        for (int aIdx = 0; aIdx < m_potentialAspectsSorted.size(); aIdx++) {
            final ConceptGraphObject aspect = m_potentialAspectsSorted.get(aIdx);
            final double aInDegree = aspect.getInDegree();
            final double existingMinusNotExisting = aInDegree - (m_members.size() - aInDegree);
            if (existingMinusNotExisting >= 0) {
                final Set<String> aNeighbors = m_aspectNeighbors.get(aspect.getId());
                //this aspect improves the score since it is an aspect of at least half of the members
                for (String memberId : aNeighbors) {
                    ConceptGraphObject member = m_members.get(memberId);
                    if (member != null) {
                        //add the aspect to its neighbors
                        member.addCGNeighbor(aspect);
                    }
                }
                //since we remove the potential aspect from the sorted potential aspect list we have to
                //decrease also the counter by one
                addAspect(aIdx--, aspect);
                //update the concept graph score
                m_score += existingMinusNotExisting * aInDegree / aspect.getDegree();
            }
        }
    }

    private void addMembers() {
        final BitSet aNeighborIdxs = new BitSet(m_aspects.size());
        //we have to check the length of the sorted members in each round since it might have changed
        for (int mIdx = 0; mIdx < m_potentialMembersSorted.size(); mIdx++) {
            final ConceptGraphObject member = m_potentialMembersSorted.get(mIdx);
            double cueValDiff = 0;
            double mScore = 0;
            aNeighborIdxs.clear();
            final Set<String> mNeighbors = m_memberNeighbors.get(member.getId());
            for (int idx = 0, length = m_aspects.size(); idx < length; idx++) {
                final ConceptGraphObject aspect = m_aspects.get(idx);
                final double aInDegree = aspect.getInDegree();
                final double aDegree = aspect.getDegree();
                if (mNeighbors.contains(aspect.getId())) {
                    //the member is a neighbor of the aspect
                    mScore += (aInDegree + 1) / aDegree;
                    if (m_useCueValidity) {
                        //this is the cue validity diff of all members that possess the aspect minus the members
                        //that do not possess the aspect
                        //the cue validity affects only the aspects that are neighbors of the new member
                        cueValDiff += (aInDegree - (m_members.size() - aInDegree)) / aDegree;
                    }
                    aNeighborIdxs.set(idx);
                } else {
                    //the aspect is no member of the neighbor so we do not need to compute the cue validity diff
                    mScore -= aInDegree / aDegree;
                }
            }
            final double memberScore = mScore + cueValDiff;
            if (memberScore >= 0) {
                //loop over all aspects that are affected by the current member
                for (int idx = aNeighborIdxs.nextSetBit(0); idx >= 0; idx = aNeighborIdxs.nextSetBit(idx + 1)) {
                    final ConceptGraphObject aspect = m_aspects.get(idx);
                    aspect.addCGNeighbor(member);
                }
                //since we remove the member from the sorted member list we have to decrease also the counter by one
                addMember(mIdx--, member);
                //update the concept graph score
                m_score += memberScore;
            }
        }
    }

    /**
     * @param idx
     * @param aspect
     */
    private void removeAspect(final int idx, final ConceptGraphObject aspect) {
        m_changed = m_potentialAspectsSorted.add(aspect);
        m_aspects.remove(idx);
//        if (LOGGER.isEnabledFor(LEVEL.INFO)) {
//            LOGGER.info(getId() + ": " + "Aspect removed: " + aspect.getId());
//        }
    }

    /**
     * @param aspect the aspect to add to the best aspects map
     */
    private void addAspect(final int idx, final ConceptGraphObject aspect) {
        m_changed = m_aspects.add(aspect);
        m_potentialAspectsSorted.remove(idx);
//        if (LOGGER.isEnabledFor(LEVEL.INFO)) {
//            LOGGER.info(getId() + ": " + "Additional aspect added: " + aspect.getId());
//        }
    }

    /**
     * @param member the member to add to the best members map
     */
    private void addMember(final int idx, final ConceptGraphObject member) {
        m_changed = m_members.put(member.getId(), member) == null;
        m_potentialMembersSorted.remove(idx);
//        if (LOGGER.isEnabledFor(LEVEL.INFO)) {
//            LOGGER.info(getId() + ": " + "Additional member added: " + member.getId());
//        }
    }

    /**
     * @param member
     */
    private void removeMember(final ConceptGraphObject member) {
        m_changed = m_potentialMembersSorted.add(member);
        m_members.remove(member.getId());
//        if (LOGGER.isEnabledFor(LEVEL.INFO)) {
//            LOGGER.info(getId() + ": " + "Member removed: " + member.getId());
//        }
    }

    private List<ConceptGraphObject> createList(final Map<String, Set<String>> candidates, final Set<String> filterIds,
        final Set<String> base, final Comparator<ConceptGraphObject> comparator, final boolean isAspect) {
        final List<ConceptGraphObject> candidateList =
                new ArrayList<ConceptGraphObject>(candidates.size() - filterIds.size());
        for (Entry<String, Set<String>> candidate : candidates.entrySet()) {
            if (!filterIds.contains(candidate.getKey())) {
                final Set<String> neighbors = candidate.getValue();
                final int inDegree = GraphUtil.computeOverlap(base, neighbors);
                final int outDegree = neighbors.size() - inDegree;
                final ConceptGraphObject object =
                        new ConceptGraphObject(candidate.getKey(), inDegree, outDegree, isAspect);
                candidateList.add(object);
            }
        }
        if (comparator != null) {
            Collections.sort(candidateList, comparator);
        }
        return candidateList;
    }
}
