/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 21.05.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import edu.cmu.koetter.util.AspectComparator;
import edu.cmu.koetter.util.MemberComparator;
import edu.cmu.koetter.util.NeighborBiGraphExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.NodeLogger;
import org.knime.core.util.Pair;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.util.GraphUtil;

/**
 *
 * @author Tobias Koetter
 */
public class IterativeFastAspectBasedConceptGraphExtractor extends FastAspectBasedConceptGraphExtractor {

    private static final NodeLogger LOGGER = NodeLogger.getLogger(IterativeFastAspectBasedConceptGraphExtractor.class);

    private static final int MAX_ITERATION = 1000;

    private Map<String, Set<String>> m_aspectNeighbors;
    private Map<String, Set<String>> m_memberNeighbors;
    private boolean m_changed = false;
    private double m_score = 0;
    private Map<String, ConceptGraphObject> m_members;
    private List<ConceptGraphObject> m_aspects;
    private List<ConceptGraphObject> m_potentialAspectsSorted;
    private List<ConceptGraphObject> m_potentialMembersSorted;

    private boolean m_useCueValidity;


    /**
     * @param view the {@link KPartiteGraphView}
     * @param scorer the {@link Scorer}
     * @param minNodeDegree the minimum node degree
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     * @param startNode the start node
     */
    public IterativeFastAspectBasedConceptGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
        final Scorer scorer, final int minNodeDegree, final Map<String, NodeNeighbors> neighborMap,
        final PersistentObject startNode) {
        super(view, scorer, minNodeDegree, neighborMap, startNode);
        m_useCueValidity = scorer.useCueValidity();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected ConceptGraph createConceptGraph(final ExecutionMonitor exec, final Collection<ConceptGraphObject> aspects,
        final Collection<ConceptGraphObject> members, final double score)
                throws CanceledExecutionException {
        final long startTime = System.currentTimeMillis();
        final NeighborBiGraphExtractor extractor = getExtractor();
        m_aspectNeighbors = extractor.getAspectNeighbors();
        m_memberNeighbors = extractor.getMemberNeighbors();
        final Set<String> aspectIds = ConceptGraphObject.createIdSet(aspects);
        final Set<String> memberIds = ConceptGraphObject.createIdSet(members);
        m_potentialMembersSorted =
                createList(m_memberNeighbors, memberIds, aspectIds, MemberComparator.getInstance(), false);
        m_potentialAspectsSorted =
                createList(m_aspectNeighbors, aspectIds, memberIds, AspectComparator.getInstance(), true);
        m_score = score;
        m_members = new HashMap<String, ConceptGraphObject>(members.size());
        for (ConceptGraphObject member : members) {
            m_members.put(member.getId(), member);
        }
        m_aspects = new ArrayList<ConceptGraphObject>(aspects);
        m_changed = !m_potentialAspectsSorted.isEmpty() || !m_potentialMembersSorted.isEmpty();
        int iterationCounter = 0;
        while (m_changed && iterationCounter < MAX_ITERATION) {
            exec.setMessage("improvement iteration: " + ++iterationCounter);
            exec.checkCanceled();
            m_changed = false;
            if (!m_potentialAspectsSorted.isEmpty()) {
                addAspects();
            }
            if (!m_potentialMembersSorted.isEmpty()) {
                addMembers();
            }
            if (removeMembers()) {
                Collections.sort(m_potentialMembersSorted, MemberComparator.getInstance());
            }
            if (removeAspects()) {
                Collections.sort(m_potentialAspectsSorted, AspectComparator.getInstance());
            }
        }
        checkForDuplicates();
        final long endTime = System.currentTimeMillis();
        final long durationTime = endTime - startTime;
        LOGGER.debug(getStartNode().getId() + ": Time for improvement over " + iterationCounter
            + " iterations: " + durationTime + " ms");
        return super.createConceptGraph(exec, m_aspects, m_members.values(), m_score);
    }

    private void checkForDuplicates() {
        final Collection<Pair<ConceptGraphObject, Double>> aspects2remove =
                new LinkedList<Pair<ConceptGraphObject, Double>>();
        for (ConceptGraphObject aspect : m_aspects) {
            final ConceptGraphObject member = m_members.get(aspect.getId());
            if (member != null) {
                final double aspectScore = computeAspectScore(aspect);
                final BitSet aNeighborIdxs = new BitSet(m_aspects.size());
                final double memberScore = computeMemberScore(aNeighborIdxs, member);
                if (memberScore < aspectScore) {
                    removeMember(aNeighborIdxs, member, memberScore);
                } else {
                    aspects2remove.add(new Pair<ConceptGraphObject, Double>(aspect, Double.valueOf(aspectScore)));
                }
            }
        }
        for (Pair<ConceptGraphObject, Double> pair : aspects2remove) {
            removeAspect(pair.getFirst(), pair.getSecond().doubleValue());
        }
    }

    private double computeMemberScore(final BitSet aNeighborIdxs, final ConceptGraphObject member) {
        //these are the aspects the member possesses
        final Set<String> memberAspects = m_memberNeighbors.get(member.getId());
        double cueValDiff = 0;
        double mScore = 0;
        for (int idx = 0, length = m_aspects.size(); idx < length; idx++) {
            final ConceptGraphObject aspect = m_aspects.get(idx);
            if (memberAspects.contains(aspect.getId())) {
                //the member possesses the current aspect thus we have to increase the score
                final double aInDegree = aspect.getInDegree();
                final double aDegree = aspect.getDegree();
                //the member is a neighbor of the aspect so the cue validity of the aspect is increased by an
                //additional member within the concept graph that is why we add 1 to the aInDegree
                mScore += (aInDegree + 1) / aDegree;
                if (m_useCueValidity) {
                    //this is the cue validity diff of all members that possess the aspect minus the members
                    //that do not possess the aspect
                    //the cue validity affects only the aspects that are neighbors of the new member
                    cueValDiff += (aInDegree - (m_members.size() - aInDegree)) / aDegree;
                }
                aNeighborIdxs.set(idx);
            } else {
                //the aspect is no member of the neighbor so we do not need to compute the cue validity diff
                mScore -= aspect.getCueValidity();
            }
        }
        final double memberScore = mScore + cueValDiff;
        return memberScore;
    }

    private double computeAspectScore(final ConceptGraphObject aspect) {
        final int aInDegree = aspect.getInDegree();
        if (aInDegree == 0) {
            //return a negative score if the aspect is not connected to any member of the concept graph
            return -1;
        }
        //this is the number of aspect neighbors that are members of the concept graph
        //minus the number members that are not neighbors of the aspect
        final int existingMinusNotExisting = aInDegree - (m_members.size() - aInDegree);
        final double aspectScore = existingMinusNotExisting * aspect.getCueValidity();
        return aspectScore;
    }

    private void addAspects() {
        //we have to check the length of the sorted aspects in each round since it might have changed
        for (int aIdx = 0; aIdx < m_potentialAspectsSorted.size(); aIdx++) {
            final ConceptGraphObject aspect = m_potentialAspectsSorted.get(aIdx);
            final double aspectScore = computeAspectScore(aspect);
            if (aspectScore >= 0) {
                addAspect(aspect, aspectScore);
                //since we remove the potential aspect from the sorted potential aspect list we have to
                //decrease also the counter by one
                aIdx--;
            }
        }
    }

    private void addMembers() {
        final BitSet aNeighborIdxs = new BitSet(m_aspects.size());
        //we have to check the length of the sorted members in each round since it might have changed
        for (int mIdx = 0; mIdx < m_potentialMembersSorted.size(); mIdx++) {
            final ConceptGraphObject member = m_potentialMembersSorted.get(mIdx);
            aNeighborIdxs.clear();
            final double memberScore = computeMemberScore(aNeighborIdxs, member);
            if (memberScore >= 0) {
                addMember(aNeighborIdxs, member, memberScore);
                //since we remove the member from the sorted member list we have to decrease also the counter by one
                mIdx--;
            }
        }
    }

    /**
     * @param aspect the aspect to add to the best aspects map
     * @param aspectScore
     */
    private void addAspect(final ConceptGraphObject aspect, final double aspectScore) {
        final Set<String> aNeighbors = m_aspectNeighbors.get(aspect.getId());
        //this aspect improves the score since it is an aspect of at least half of the members
        for (String memberId : aNeighbors) {
            ConceptGraphObject member = m_members.get(memberId);
            if (member != null) {
                //add the aspect to its neighbors
                member.addCGNeighbor(aspect);
            }
        }
        m_changed = m_aspects.add(aspect);
        m_potentialAspectsSorted.remove(aspect);
        //update the concept graph score
        m_score += aspectScore;
    }

    private void addMember(final BitSet aNeighborIdxs, final ConceptGraphObject member,
        final double memberScore) {
        //loop over all aspects that are affected by the current member
        for (int idx = aNeighborIdxs.nextSetBit(0); idx >= 0; idx = aNeighborIdxs.nextSetBit(idx + 1)) {
            final ConceptGraphObject aspect = m_aspects.get(idx);
            aspect.addCGNeighbor(member);
        }
        m_changed = m_members.put(member.getId(), member) != null;
        m_potentialMembersSorted.remove(member);
        //update the concept graph score
        m_score += memberScore;
    }

    private boolean removeAspects() {
      //we have to check the length of the sorted aspects in each round since it might have changed
        boolean changed = false;
        for (int aIdx = 0; aIdx < m_aspects.size(); aIdx++) {
            final ConceptGraphObject aspect = m_aspects.get(aIdx);
            final double aspectScore = computeAspectScore(aspect);
            if (aspectScore <= 0) {
                removeAspect(aspect, aspectScore);
                //since we remove the potential aspect from the sorted potential aspect list we have to
                //decrease also the counter by one
                aIdx--;
                changed = true;
            }
        }
        return changed;
    }

    private boolean removeMembers() {
        boolean changed = false;
        final BitSet aNeighborIdxs = new BitSet(m_aspects.size());
        final List<ConceptGraphObject> members = new ArrayList<ConceptGraphObject>(m_members.values());
        //we have to check the length of the sorted members in each round since it might have changed
        for (int mIdx = 0; mIdx < members.size(); mIdx++) {
            final ConceptGraphObject member = members.get(mIdx);
            aNeighborIdxs.clear();
            final double memberScore = computeMemberScore(aNeighborIdxs, member);
            if (memberScore <= 0) {
                changed = true;
                removeMember(aNeighborIdxs, member, memberScore);
                members.remove(mIdx--);
            }
        }
        return changed;
    }

    private void removeAspect(final ConceptGraphObject aspect, final double aspectScore) {
        final Set<String> aNeighbors = m_aspectNeighbors.get(aspect.getId());
        //this aspect decreases the score since it is an aspect with less than half of the members
        for (String memberId : aNeighbors) {
            ConceptGraphObject member = m_members.get(memberId);
            if (member != null) {
                //add the aspect to its neighbors
                member.removeCGNeighbor(aspect);
            }
        }
        m_changed = m_potentialAspectsSorted.add(aspect);
        m_aspects.remove(aspect);
        //update the concept graph score
        m_score -= aspectScore;
    }

    private void removeMember(final BitSet aNeighborIdxs, final ConceptGraphObject member, final double memberScore) {
        //loop over all aspects that are affected by the current member
        for (int idx = aNeighborIdxs.nextSetBit(0); idx >= 0; idx = aNeighborIdxs.nextSetBit(idx + 1)) {
            final ConceptGraphObject aspect = m_aspects.get(idx);
            aspect.removeCGNeighbor(member);
        }
        //since we remove the member from the sorted member list we have to decrease also the counter by one
        m_changed = m_potentialMembersSorted.add(member);
        m_members.remove(member.getId());
        //update the concept graph score
        m_score -= memberScore;
    }

    private List<ConceptGraphObject> createList(final Map<String, Set<String>> candidates, final Set<String> filterIds,
        final Set<String> base, final Comparator<ConceptGraphObject> comparator, final boolean isAspect) {
        final List<ConceptGraphObject> candidateList =
                new ArrayList<ConceptGraphObject>(candidates.size() - filterIds.size());
        for (Entry<String, Set<String>> candidate : candidates.entrySet()) {
            if (!filterIds.contains(candidate.getKey())) {
                final Set<String> neighbors = candidate.getValue();
                final int inDegree = GraphUtil.computeOverlap(base, neighbors);
                final int outDegree = neighbors.size() - inDegree;
                final ConceptGraphObject object =
                        new ConceptGraphObject(candidate.getKey(), inDegree, outDegree, isAspect);
                candidateList.add(object);
            }
        }
        if (comparator != null) {
            Collections.sort(candidateList, comparator);
        }
        return candidateList;
    }
}
