/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 29.04.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

/**
 *
 * @author Tobias Koetter
 */
public class ScoreCache implements Iterable<ConceptGraphObject> {

        private final List<ConceptGraphObject> m_objects;
        private int m_startIdx = 0;
        private boolean m_sorted = false;
        private double m_sum = 0;
        private double m_product = 1;
        private final Comparator<ConceptGraphObject> m_comp;

        /**
         * @param size the initial size of the cache
         */
        public ScoreCache(final int size) {
            this(size, ConceptGraphObject.ASC_SCORE_COMP);
        }

        /**
         * @param size the initial size of the cache
         * @param comp the {@link Comparator} to use
         */
        public ScoreCache(final int size, final Comparator<ConceptGraphObject> comp) {
            m_objects  = new ArrayList<ConceptGraphObject>(size);
            m_comp = comp;
        }

        /**
         *
         */
        public void reset() {
            m_startIdx = 0;
            m_objects.clear();
            m_sorted = false;
            m_sum = 0;
            m_product = 1;
        }

        /**
         * @param object the {@link ConceptGraphObject}
         */
        public void add(final ConceptGraphObject object) {
//            LOGGER.info("->" + object.getId() + ": " + score + ":" + object.isAspect());
            m_sum += object.getScore();
            m_product *= object.getScore();
            m_objects.add(object);
        }

        /**
         * Removes the current minimum object.
         * @return the current minimum object
         */
        public ConceptGraphObject removeObject() {
            final ConceptGraphObject object = getObject();
            final double score = object.getScore();
            m_sum -= score;
            if (score != 0) {
                m_product /= score;
            }
            m_startIdx++;
            return object;
        }

        /**
         * @return the sum of the scores
         */
        public double getScoreSum() {
            return m_sum;
        }

        /**
         * @return the product of the scores
         */
        public double getScoreProduct() {
            return m_product;
        }

        /**
         * @return the minScore
         */
        public double getScore() {
            final ConceptGraphObject object = getObject();
            if (object == null) {
                return 0;
            }
            return object.getScore();
        }

        /**
         * @return the secondMinScore
         */
        public double getSecondScore() {
            final ConceptGraphObject secondObject = getSecondObject();
            if (secondObject == null) {
                return 0;
            }
            return secondObject.getScore();
        }

        /**
         * @return the minObject
         */
        public ConceptGraphObject getObject() {
            return getObject(m_startIdx);
        }

        /**
         * @return the secondObject
         */
        public ConceptGraphObject getSecondObject() {
            return getObject(m_startIdx + 1);
        }

        /**
         * @param idx
         * @return
         */
        private ConceptGraphObject getObject(final int idx) {
            sort();
            if (idx >= m_objects.size()) {
                return null;
            }
            return m_objects.get(idx);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public String toString() {
            StringBuilder builder = new StringBuilder();
            builder.append("ScoreCache [");
            if (m_objects != null) {
                builder.append("m_objects=");
                builder.append(m_objects);
                builder.append(", ");
            }
            builder.append("m_startIdx=");
            builder.append(m_startIdx);
            builder.append(", m_sorted=");
            builder.append(m_sorted);
            builder.append(", m_sum=");
            builder.append(m_sum);
            builder.append(", ");
            if (m_comp != null) {
                builder.append("m_comp=");
                builder.append(m_comp);
            }
            builder.append("]");
            return builder.toString();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public Iterator<ConceptGraphObject> iterator() {
            return new Iterator<ConceptGraphObject>() {

                private int m_idx = m_startIdx;

                @Override
                public boolean hasNext() {
                    return m_idx < m_objects.size();
                }

                @Override
                public ConceptGraphObject next() {
                    return getObject(m_idx++);
                }

                @Override
                public void remove() {
                    throw new UnsupportedOperationException();
                }
            };
        }

        /**
         * @return the number of objects
         */
        public int size() {
            return m_objects.size() - m_startIdx;
        }

        /**
         * @return all added objects
         */
        public Collection<ConceptGraphObject> getObjects() {
            if (m_startIdx == 0) {
                return m_objects;
            }
            final ArrayList<ConceptGraphObject> objects = new ArrayList<ConceptGraphObject>(size());
            Iterator<ConceptGraphObject> iter = iterator();
            while (iter.hasNext()) {
                objects.add(iter.next());
            }
            return objects;
        }

        /**
         * Sorts the objects.
         */
        public void sort() {
            if (!m_sorted && m_comp != null) {
                Collections.sort(m_objects, m_comp);
                m_sorted = true;
            }
        }

        /**
         * @return <code>true</code> if the {@link ScoreCache} contains no more {@link ConceptGraphObject}s
         */
        public boolean isEmpty() {
            return size() > 0;
        }
    }