/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 19.04.2013 by koetter
 */
package edu.cmu.koetter.conceptgraph;

import java.util.Comparator;


/**
 *
 * @author koetter
 * @param <O> the object type
 */
public class ValuedObject<O extends Object> implements Comparable<ValuedObject<O>> {

    private static final class ValuedObjectComparator implements Comparator<ValuedObject<?>> {

        private final boolean m_descending;

        private ValuedObjectComparator(final boolean descending) {
            m_descending = descending;
        }
        /**
         * {@inheritDoc}
         */
        @Override
        public int compare(final ValuedObject<?> o1, final ValuedObject<?> o2) {
            if (o1 == o2) {
                return 0;
            }
            if (o1 == null) {
                return -1;
            }
            if (o2 == null) {
                return 1;
            }
            final double v1 = o1.getValue();
            final double v2 = o2.getValue();
            int compare = Double.compare(v1, v2);
            if (compare == 0) {
                //if the value is the same compare the string representation of the objects
                compare = o1.getObject().toString().compareTo(o2.getObject().toString());
            }
            if (m_descending) {
                compare *= -1;
            }
            return compare;
        }
    }

    /**Descending comparator.*/
    public static final ValuedObjectComparator DESC_COMPARATOR = new ValuedObjectComparator(true);
    /**Ascending comparator.*/
    public static final ValuedObjectComparator ASC_COMPARATOR = new ValuedObjectComparator(false);

    private final O m_object;
    private final double m_value;

    /**
     * @param object the object
     * @param value its value
     */
    public ValuedObject(final O object, final double value) {
        if (object == null) {
            throw new IllegalArgumentException("object must not be null");
        }
        m_object = object;
        m_value = value;
    }

    /**
     * @return the accumulated activation value
     */
    public double getValue() {
        return m_value;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int compareTo(final ValuedObject<O> o) {
        if (o == null) {
            return -1;
        }
        int compare = Double.compare(o.getValue(), getValue());
        if (compare == 0) {
            //if the value is the same compare the string representation of the objects
            return m_object.toString().compareTo(o.toString());
        }
        return compare;
    }

    /**
     * @return the object
     */
    public O getObject() {
        return m_object;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();
        builder.append("ValuedObject [");
            builder.append(m_object);
            builder.append("=");
        builder.append(m_value);
        builder.append("]");
        return builder.toString();
    }


}
