package edu.cmu.koetter.linkprediction;
/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 03.07.2013 by koetter
 */


import edu.cmu.koetter.conceptgraph.ConceptGraphExtractorNodeModel;
import edu.cmu.koetter.conceptgraph.ConceptGraphObject;
import edu.cmu.koetter.linkprediction.MissingLinks.MissingLink;
import edu.cmu.koetter.util.NeighborExtractor;
import edu.cmu.koetter.util.NodeNeighbors;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;
import java.util.concurrent.Semaphore;
import org.knime.core.data.DataCell;
import org.knime.core.data.DataColumnSpec;
import org.knime.core.data.DataColumnSpecCreator;
import org.knime.core.data.DataRow;
import org.knime.core.data.DataTableSpec;
import org.knime.core.data.RowKey;
import org.knime.core.data.collection.CollectionDataValue;
import org.knime.core.data.collection.ListCell;
import org.knime.core.data.def.DefaultRow;
import org.knime.core.data.def.DoubleCell;
import org.knime.core.data.def.LongCell;
import org.knime.core.data.def.StringCell;
import org.knime.core.node.BufferedDataContainer;
import org.knime.core.node.BufferedDataTable;
import org.knime.core.node.ExecutionContext;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.InvalidSettingsException;
import org.knime.core.node.KNIMEConstants;
import org.knime.core.node.NodeSettingsRO;
import org.knime.core.node.NodeSettingsWO;
import org.knime.core.node.defaultnodesettings.SettingsModelInteger;
import org.knime.core.node.defaultnodesettings.SettingsModelIntegerBounded;
import org.knime.core.node.defaultnodesettings.SettingsModelString;
import org.knime.core.util.ThreadPool;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.util.GraphUtil;
import org.knime.network.core.knime.node.KPartiteGraphViewAndTable2TableNodeModel;
import org.knime.network.core.knime.port.GraphPortObjectSpec;

/**
 * Node model of the missing link predictor node.
 * @author Tobias Koetter
 */
public class LinkPredictorNodeModel extends KPartiteGraphViewAndTable2TableNodeModel {

    private final SettingsModelString m_conceptCol = createConceptColModel();

    private final SettingsModelString m_aspectCol = createAspectColModel();

    private final SettingsModelString m_memberCol = createMemberColModel();

    private final SettingsModelInteger m_maxNoOfProcesses = createNoOfProcessesModel();

    private static final DataCell ASPECT_TYPE_CELL = new StringCell("A");
    private static final DataCell CONCEPT_TYPE_CELL = new StringCell("C");

    /**
     * @return the maximum number of processes model
     */
    static SettingsModelInteger createNoOfProcessesModel() {
        return new SettingsModelIntegerBounded("maximumNoOfProcesses",
                                           KNIMEConstants.GLOBAL_THREAD_POOL.getMaxThreads(), 1, Integer.MAX_VALUE);
    }

    /**
     * @return the aspect column model
     */
    static SettingsModelString createAspectColModel() {
        return new SettingsModelString("aspectColumnName", null);
    }

    /**
     * @return the concept column model
     */
    static SettingsModelString createConceptColModel() {
        return new SettingsModelString("conceptColumnName", null);
    }
    /**
     * @return the member column model
     */
    static SettingsModelString createMemberColModel() {
        return new SettingsModelString("memberColumnName", null);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected BufferedDataTable execute(final ExecutionContext exec,
        final KPartiteGraphView<PersistentObject, Partition> view, final BufferedDataTable table) throws Exception {
        final int conceptColIdx = table.getSpec().findColumnIndex(m_conceptCol.getStringValue());
        if (conceptColIdx < 0) {
            throw new InvalidSettingsException("Concept column not found");
        }
        final int aspectColIdx = table.getSpec().findColumnIndex(m_aspectCol.getStringValue());
        if (aspectColIdx < 0) {
            throw new InvalidSettingsException("Aspect column not found");
        }
        final int memberColIdx = table.getSpec().findColumnIndex(m_memberCol.getStringValue());
        if (memberColIdx < 0) {
            throw new InvalidSettingsException("Member column not found");
        }
        exec.setMessage("Caching neighbors");
        final int procCount = m_maxNoOfProcesses.getIntValue();
        final NeighborExtractor extractor = new NeighborExtractor(view);
        //extract the neighbors of all nodes to speedup the node
        final Map<String, NodeNeighbors> neighborMap =
                extractor.getNodeNeighbors(exec.createSubProgress(0.05), procCount);
        //initialize the thread pool
        final ThreadPool pool = KNIMEConstants.GLOBAL_THREAD_POOL.createSubPool();
        //The semaphore restricts the number of concurrent processes
        final Semaphore semaphore = new Semaphore(procCount);
        if (procCount == 1) {
            //acquire the only permit to prevent memory problems
            semaphore.acquire();
        }
        exec.setMessage("Processing input table...");
        final Queue<Future<MissingLinks>> futures = new LinkedList<Future<MissingLinks>>();
        try {
            final int noOfRows = table.getRowCount();
            for (final DataRow row : table) {
                final ExecutionMonitor subExec = exec.createSubProgress(0.05 / noOfRows);
                exec.checkCanceled();
                final Collection<String> conceptIds = extractIds(conceptColIdx, row);
                if (conceptIds == null) {
                    continue;
                }
                final Collection<String> aspectIds = extractIds(aspectColIdx, row);
                if (aspectIds == null) {
                    continue;
                }
                final Collection<String> memberIds = extractIds(memberColIdx, row);
                if (memberIds == null) {
                    continue;
                }
                futures.add(pool.enqueue(createFuture(subExec, view, neighborMap, conceptIds, aspectIds, memberIds)));
            }
            if (procCount == 1) {
                //release the only permit to start the Runnables
                semaphore.release();
            }
            exec.setMessage("Writing output table...");
            final BufferedDataContainer dc = exec.createDataContainer(createSpec());
            final ExecutionContext subExec = exec.createSubExecutionContext(0.9);
            int counter = 0;
            int rowCounter = 0;
            while (!futures.isEmpty()) {
                //get AND remove the finished future from the queue
                final Future<MissingLinks> future = futures.poll();
                subExec.checkCanceled();
                final MissingLinks links = future.get();
                final DataCell conceptIdsCell = ConceptGraphExtractorNodeModel.createListCell(links.getConceptIds());
                final DataCell sizeCell = new DoubleCell(links.getSize());
                for (final MissingLink link : links.getMissingLinks()) {
                    final List<DataCell> cells = new LinkedList<DataCell>();
                    cells.add(conceptIdsCell);
                    switch (link.getType()) {
                        case ASPECT:
                            cells.add(ASPECT_TYPE_CELL);
                            break;
                        case CONCEPT:
                            cells.add(CONCEPT_TYPE_CELL);
                            break;
                    }
                    cells.add(new StringCell(link.getAspectId()));
                    cells.add(new StringCell(link.getMemberId()));
                    cells.add(new LongCell(link.getAspectDegree()));
                    cells.add(new DoubleCell(link.getConfidence()));
                    cells.add(new DoubleCell(link.getMemberOverlap()));
                    cells.add(new DoubleCell(link.getAspectOverlap()));
                    cells.add(new DoubleCell(link.getAspectCueValidity()));
                    cells.add(sizeCell);
                    final DataRow row = new DefaultRow(RowKey.createRowKey(rowCounter++), cells);
                    dc.addRowToTable(row);
                }
                counter++;
                subExec.setProgress(counter / (double) noOfRows, counter + " of " + noOfRows + " rows analyzed");
            }
            dc.close();
            return dc.getTable();
        } catch (final Exception e) {
            pool.interruptAll();
            throw e;
        }
    }

    /**
     * @return
     */
    private DataTableSpec createSpec() {
        final Collection<DataColumnSpec> specs = new LinkedList<DataColumnSpec>();
        final DataColumnSpecCreator creator = new DataColumnSpecCreator(ConceptGraphExtractorNodeModel.COL_CONCEPT_IDS,
            ListCell.getCollectionType(StringCell.TYPE));
        specs.add(creator.createSpec());
        creator.setName("Type");
        creator.setType(StringCell.TYPE);
        specs.add(creator.createSpec());
        creator.setName("Aspect id");
        specs.add(creator.createSpec());
        creator.setName("Member id");
        specs.add(creator.createSpec());
        creator.setName("Aspect degree");
        creator.setType(LongCell.TYPE);
        specs.add(creator.createSpec());
        creator.setName("Confidence");
        creator.setType(DoubleCell.TYPE);
        specs.add(creator.createSpec());
        creator.setName("Member overlap");
        specs.add(creator.createSpec());
        creator.setName("Aspect overlap");
        specs.add(creator.createSpec());
        creator.setName("Aspect cue validity");
        specs.add(creator.createSpec());
        creator.setName("Concept graph size");
        specs.add(creator.createSpec());
        return new DataTableSpec(specs.toArray(new DataColumnSpec[0]));
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected DataTableSpec getTableSpec(final GraphPortObjectSpec viewSpec, final DataTableSpec tableSpec)
        throws InvalidSettingsException {
        if (m_conceptCol.getStringValue() == null) {
            boolean found = false;
            if (tableSpec.containsName(ConceptGraphExtractorNodeModel.COL_CONCEPT_IDS)) {
                m_conceptCol.setStringValue(ConceptGraphExtractorNodeModel.COL_CONCEPT_IDS);
                found = true;
            }
            if (tableSpec.containsName(ConceptGraphExtractorNodeModel.COL_ASPECT_IDS)) {
                m_aspectCol.setStringValue(ConceptGraphExtractorNodeModel.COL_ASPECT_IDS);
                found = true;
            }
            if (tableSpec.containsName(ConceptGraphExtractorNodeModel.COL_MEMBER_IDS)) {
                m_memberCol.setStringValue(ConceptGraphExtractorNodeModel.COL_MEMBER_IDS);
                found = true;
            }
            if (found) {
                setWarningMessage("Node autoconfigured");
            }
        }
        return createSpec();
    }

    private Callable<MissingLinks> createFuture(final ExecutionMonitor exec,
        final KPartiteGraphView<PersistentObject, Partition> view, final Map<String, NodeNeighbors> neighborMap,
        final Collection<String> conceptIds, final Collection<String> aspectIds, final Collection<String> memberIds) {
        return new Callable<MissingLinks>() {
            @Override
            public MissingLinks call() throws Exception {
                exec.checkCanceled();
                final Set<String> memberIdSet = new HashSet<String>(memberIds);
                final MissingLinks missingLinks = new MissingLinks(conceptIds, aspectIds, memberIds);
                int counter = 0;
//                final Set<String> potentialConceptMembers = new HashSet<String>();
                final Map<String, ConceptGraphObject> aspects =
                        new HashMap<String, ConceptGraphObject>(aspectIds.size());
//                double sumAspectCV = 0;
                for (final String aspectId : aspectIds) {
                    exec.setProgress(counter / (double) aspectIds.size());
                    exec.checkCanceled();
                    final NodeNeighbors aspectNeighbors = neighborMap.get(aspectId);
                    if (aspectNeighbors == null) {
                        throw new IllegalStateException("Could not find aspect node with id " + aspectId);
                    }
                    final Set<String> neighbors = aspectNeighbors.getPredecessors();
                    final int inDegree = GraphUtil.computeOverlap(memberIdSet, neighbors);
                    final int outDegree = neighbors.size() - inDegree;
                    final ConceptGraphObject aspectObject = new ConceptGraphObject(aspectId, inDegree, outDegree, true);
                    aspects.put(aspectId, aspectObject);
//                    sumAspectCV += aspectObject.getCueValidity();
                    if (inDegree == 0) {
                        continue;
                    }
                    final Collection<String> missingMemberIds = GraphUtil.minus(memberIdSet, neighbors);
                    missingLinks.addMissingAspectLinks(aspectObject, missingMemberIds);
//                    if (counter == 0) {
//                        potentialConceptMembers.addAll(neighbors);
//                    } else {
//                        potentialConceptMembers.retainAll(neighbors);
//                    }
                    counter++;
                }
                //remove all nodes that are already part of the concept graph
//                potentialConceptMembers.removeAll(conceptIds);
//                potentialConceptMembers.removeAll(aspectIds);
//                potentialConceptMembers.removeAll(memberIds);
//                for (final String potentialMember : potentialConceptMembers) {
//                    final NodeNeighbors potentialMemberAspects = neighborMap.get(potentialMember);
//                    final Set<String> successors = potentialMemberAspects.getSuccessors();
//                    double potentialMemberScore = 0;
//                    for (final String successor : successors) {
//                        final ConceptGraphObject aspect = aspects.get(successor);
//                        if (aspect != null) {
//                            potentialMemberScore += aspect.getCueValidity();
//                        }
//                    }
//                    missingLinks.addPotentialMember(potentialMember, potentialMemberScore);
//                }
                return missingLinks;
            }
        };
    }

    /**
     * @param colIdx
     * @param row
     * @return
     */
    private Collection<String> extractIds(final int colIdx, final DataRow row) {
        final DataCell cell = row.getCell(colIdx);
        if (cell.isMissing()) {
            return null;
        }
        if (cell instanceof CollectionDataValue) {
            final CollectionDataValue coll = (CollectionDataValue)cell;
            if (coll.size() == 0) {
                return null;
            }
            final List<String> ids = new ArrayList<String>(coll.size());
            for (final DataCell id : coll) {
                if (!id.isMissing()) {
                    ids.add(id.toString());
                }
            }
            return ids;
        }
        return null;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void resetInternal() {
        //nothing to reset
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void saveSettingsTo(final NodeSettingsWO settings) {
        m_aspectCol.saveSettingsTo(settings);
        m_conceptCol.saveSettingsTo(settings);
        m_memberCol.saveSettingsTo(settings);
        m_maxNoOfProcesses.saveSettingsTo(settings);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void validateSettings(final NodeSettingsRO settings) throws InvalidSettingsException {
        m_aspectCol.validateSettings(settings);
        m_conceptCol.validateSettings(settings);
        m_memberCol.validateSettings(settings);
        m_maxNoOfProcesses.validateSettings(settings);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void loadValidatedSettingsFrom(final NodeSettingsRO settings) throws InvalidSettingsException {
        m_aspectCol.loadSettingsFrom(settings);
        m_conceptCol.loadSettingsFrom(settings);
        m_memberCol.loadSettingsFrom(settings);
        m_maxNoOfProcesses.loadSettingsFrom(settings);
    }
}
