/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 11.04.2013 by koetter
 */
package edu.cmu.koetter.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import org.knime.core.data.DataCell;
import org.knime.core.data.collection.CollectionCellFactory;
import org.knime.core.data.def.StringCell;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.NodeLogger;
import org.knime.core.util.MutableInteger;
import org.knime.network.core.api.GraphObjectIterator;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.PartitionType;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.util.GraphUtil;

/**
 *
 * @author Tobias Koetter
 * @param <V> the value that holds the information
 */
public abstract class BiGraphExtractor<V> {
    private static final NodeLogger LOGGER = NodeLogger.getLogger(BiGraphExtractor.class);
    private String[] m_dimensions;
    private String[] m_objectIds;
    private boolean m_reverse;
    private final KPartiteGraphView<PersistentObject, Partition> m_view;
    private final EdgeWeightCalculator m_calculator;

    /**
     * @param view {@link KPartiteGraphView}
     * @param calculator the {@link EdgeWeightCalculator} to use
     */
    protected BiGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                                    final EdgeWeightCalculator calculator) {
        m_view = view;
        m_calculator = calculator;
    }

    /**
     * Starts the conversion.
     * @param exec {@link ExecutionMonitor} to provide progress
     * @param minNodeDegree minimum node degree
     * @param startNodes the start node
     * @throws PersistenceException if an exception occurs
     * @throws CanceledExecutionException if the operation has been canceled
     */
    public void run(final ExecutionMonitor exec, final int minNodeDegree,
                 final PersistentObject... startNodes) throws PersistenceException, CanceledExecutionException {
        Partition[] edgePartitions = m_view.getPartitions(PartitionType.EDGE).toArray(new Partition[0]);
        Partition[] nodePartitions = m_view.getPartitions(PartitionType.NODE).toArray(new Partition[0]);

        GraphObjectIterator<PersistentObject> conceptEdges = null;
        GraphObjectIterator<PersistentObject> memberNodes = null;
        GraphObjectIterator<PersistentObject> memberEdges = null;
        GraphObjectIterator<PersistentObject> aspectNodes = null;
        try {
            final Map<PersistentObject, Set<PersistentObject>> memberMap =
                    new HashMap<PersistentObject, Set<PersistentObject>>();
            final Map<PersistentObject, MutableInteger> aspectIds =
                    new HashMap<PersistentObject, MutableInteger>();
            for (PersistentObject startNode : startNodes) {
                conceptEdges = m_view.getIncidentEdges(startNode, edgePartitions);
                for (final PersistentObject conceptEdge : conceptEdges) {
                    final boolean isAspect = !m_view.isTarget(conceptEdge, startNode);
                    if (isAspect) {
                        //skip edges where the start node is not the target
                        continue;
                    }
                    memberNodes = m_view.getIncidentNodes(conceptEdge, nodePartitions);
                    for (PersistentObject memberNode : memberNodes) {
                        if (memberNode.equals(startNode)) {
                            continue;
                        }
                        boolean isTarget = !m_view.isSource(conceptEdge, memberNode);
                        if (isTarget) {
                            //skip edges where the member node is not the source
                            continue;
                        }
                        if (!isAspect && !isTarget) { //  && !startNodeAspects.contains(memberNode)) {
                            //the node points to the start node -> member of the start node
                            final Set<PersistentObject> memberNeighbors = new HashSet<PersistentObject>();
                            memberMap.put(memberNode, memberNeighbors);
                            memberEdges = m_view.getIncidentEdges(memberNode, edgePartitions);
                            for (PersistentObject memberEdge : memberEdges) {
                                if (!m_view.isSource(memberEdge, memberNode)) {
                                    //skip edges where the member is not the source
                                    //  (use the negation since this works also with undirected networks)
                                    continue;
                                }
                                aspectNodes = m_view.getIncidentNodes(memberEdge, nodePartitions);
                                for (PersistentObject aspectNode : aspectNodes) {
                                    if (aspectNode.equals(memberNode) //|| memberMap.containsKey(aspectNode)
                                            || !m_view.isTarget(memberEdge, aspectNode)) {
                                        //skip aspect nodes that are already classified as members or that are
                                        //not the targets of the member edge
                                        //  (use the negation since this works also with undirected networks)
                                        continue;
                                    }
                                    //add the aspect to the current member
                                    memberNeighbors.add(aspectNode);
                                    MutableInteger counter = aspectIds.get(aspectNode);
                                    if (counter == null) {
                                        counter = new MutableInteger(0);
                                        aspectIds.put(aspectNode, counter);
                                    }
                                    counter.inc();
                                }
                                exec.checkCanceled();
                                GraphUtil.close(aspectNodes);
                            }
                            GraphUtil.close(memberEdges);
                        }
                        GraphUtil.close(memberNodes);
                    }
                    GraphUtil.close(conceptEdges);
                }
            }

            //remove all members from the aspects
            for (PersistentObject member : memberMap.keySet()) {
//                MutableInteger counter = aspectIds.get(member);
//                if (counter == null || counter.intValue() < memberMap.size() / 2) {
                    aspectIds.remove(member);
//                }
            }
            //remove all aspects with a lower in degree then the given minimum in degree
            final Collection<PersistentObject> toRemove = new LinkedList<PersistentObject>();
            for (Entry<PersistentObject, MutableInteger> entry : aspectIds.entrySet()) {
                if (entry.getValue().intValue() < minNodeDegree) {
                    toRemove.add(entry.getKey());
                }
            }
            if (!toRemove.isEmpty()) {
                final String id = GraphUtil.convertToString(null, null, ", ", Arrays.asList(startNodes));
                LOGGER.info(id + ": " + toRemove.size() + " nodes with degree < " + minNodeDegree + " removed ("
                            + toRemove.size() / (double) aspectIds.size() * 100 + "%)");
                for (PersistentObject object : toRemove) {
                    aspectIds.remove(object);
                }
            }

            m_dimensions = new String[aspectIds.size()];
            init(memberMap.size(), aspectIds.size(), startNodes);
            final Map<PersistentObject, Integer> aspectIdMap = new HashMap<PersistentObject, Integer>(aspectIds.size());
            for (PersistentObject aspect : aspectIds.keySet()) {
                int index = aspectIdMap.size();
                aspectIdMap.put(aspect, new Integer(index));
                m_dimensions[index] = aspect.getId();
                //remove all remaining aspects from the members
                memberMap.remove(aspect);
            }


            int rowIdx = 0;
            m_objectIds = new String[memberMap.size()];
            for (Entry<PersistentObject, Set<PersistentObject>> entry : memberMap.entrySet()) {
                PersistentObject member = entry.getKey();
                Set<PersistentObject> aspects = entry.getValue();
                m_objectIds[rowIdx] = member.getId();
                for (PersistentObject aspect : aspects) {
                    Integer idx = aspectIdMap.get(aspect);
                    if (idx != null) {
                        final double weight =
                                m_calculator.calc(m_view, aspectIdMap.keySet(), memberMap.keySet(), aspect, member);
                        createEdge(rowIdx, member, idx.intValue(), aspect, weight);
                    }
                }
                rowIdx++;
            }
            finish(exec);
        } finally {
            GraphUtil.close(conceptEdges);
            GraphUtil.close(memberNodes);
            GraphUtil.close(memberEdges);
            GraphUtil.close(aspectNodes);
        }
    }

    /**
     * Called once the extractor is no longer needed. Can be used to free resources.
     * @throws PersistenceException if an exception occurs
     */
    public abstract void cleanup() throws PersistenceException;

    /**
     * @param exec {@link ExecutionMonitor}
     * @throws PersistenceException if an exception occurs
     * @throws CanceledExecutionException if the operation has been canceled
     */
    protected abstract void finish(ExecutionMonitor exec)
            throws PersistenceException, CanceledExecutionException;

    /**
     * @return the view
     */
    public KPartiteGraphView<PersistentObject, Partition> getView() {
        return m_view;
    }

    /**
     * @param memberIdx the row index
     * @param member the source node
     * @param aspectIdx the column index
     * @param aspect the target node
     * @param weight the weight of the edge
     */
    protected abstract void createEdge(final int memberIdx, PersistentObject member,
                                       final int aspectIdx, PersistentObject aspect, double weight);

    /**
     * @param startNodes the start nodes
     * @param noOfMembers the number of members
     * @param noOfAspects the number of aspects
     * @throws PersistenceException if an exception occurs
     */
    protected abstract void init(final int noOfMembers, final int noOfAspects, final PersistentObject... startNodes)
            throws PersistenceException;

    /**
     * @return <code>true</code> if the dimensions represent members and objects represent aspects or
     * <code>false</code> if the dimensions represent the aspects and the objects represent the members
     */
    public boolean isReverse() {
        return m_reverse;
    }

    /**
     * @return the data
     */
    public abstract V getData();

    /**
     * @return the objectIds
     */
    private String[] getObjectIds() {
        return m_objectIds;
    }

    /**
     * @return the dimensions
     */
    private String[] getDimensions() {
        return m_dimensions;
    }

    /**
     * @return the aspects
     */
    public String[] getAspects() {
        if (isReverse()) {
            return getObjectIds();
        }
        return getDimensions();
    }

    /**
     * @return the members
     */
    public String[] getMembers() {
        if (isReverse()) {
            return getDimensions();
        }
        return getObjectIds();
    }

    /**
     * @return the number of objects
     */
    public int getMemberCount() {
        return getMembers().length;
    }

    /**
     * @return the number of dimension
     */
    public int getAspectCount() {
        return getAspects().length;
    }

    /**
     * @param objects the set of indices of the objects
     * @return the objects cell
     */
    public DataCell getMemberCell(final Set<Integer> objects) {
        return createCollectionCell(objects, getMembers());
    }

    /**
     * @param properties the property indices
     * @return the dimensions cell
     */
    public DataCell getAspectCell(final Set<Integer> properties) {
        return createCollectionCell(properties, getAspects());
    }

    private DataCell createCollectionCell(final Set<Integer> idxs, final Object[] objects) {
        final Collection<DataCell> cells = new ArrayList<DataCell>(idxs.size());
        for (Integer idx : idxs) {
            cells.add(new StringCell(objects[idx.intValue()].toString()));
        }
        return CollectionCellFactory.createSetCell(cells);
    }

}