/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 04.04.2013 by koetter
 */
package edu.cmu.koetter.util;

import edu.cmu.koetter.conceptgraph.ConceptGraph;
import java.util.Arrays;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.network.core.api.KPartiteGraph;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.GraphFactory;
import org.knime.network.core.core.PartitionType;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.util.GraphUtil;
import org.knime.network.core.knime.GraphRepository;

/**
 *
 * @author Tobias Koetter
 */
public class BiGraphGraphExtractor extends BiGraphExtractor<KPartiteGraphView<PersistentObject, Partition>> {

    private KPartiteGraph<PersistentObject, Partition> m_net;
    private final boolean m_directed;
    private Partition m_memberPartition;
    private Partition m_aspectPartition;

    /**
     * @param view the {@link KPartiteGraphView}
     * @param calculator the {@link EdgeWeightCalculator}
     * @param directed <code>true</code> if the extracted network should be directed
     */
    public BiGraphGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
                                      final EdgeWeightCalculator calculator, final boolean directed) {
        super(view, calculator);
        m_directed = directed;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void init(final int noOfRows, final int noOfCols, final PersistentObject... startNodes)
            throws PersistenceException {
        m_net = GraphFactory.createNet(GraphUtil.convertToString(null, null, ",", Arrays.asList(startNodes)));
        m_memberPartition = m_net.createPartition(ConceptGraph.MEMBERS_PARTITION, PartitionType.NODE);
        m_aspectPartition = m_net.createPartition(ConceptGraph.ASPECTS_PARTITION, PartitionType.NODE);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void finish(final ExecutionMonitor exec) throws PersistenceException, CanceledExecutionException {
        // nothing to do
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void cleanup() throws PersistenceException {
        GraphRepository.getInstance().destroy(m_net);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected void createEdge(final int sourceIdx, final PersistentObject source,
                              final int targetIdx, final PersistentObject target, final double weight) {
        final String sourceId = source.getId();
        final String targetId = target.getId();
        try {
            final String edgeId = getEdgeId(source.getId(), targetId);
            PersistentObject edge = m_net.getEdge(edgeId);
            if (edge != null) {
                //the edge already exists
                return;
            }
            final PersistentObject sourceNode = getCreateNode(sourceId, m_memberPartition);
            final PersistentObject targetNode = getCreateNode(targetId, m_aspectPartition);
            if (m_directed) {
                edge = m_net.createDirectedEdge(edgeId, sourceNode, targetNode);
            } else {
                edge = m_net.createEdge(edgeId, sourceNode, targetNode);
            }
            m_net.setWeight(edge, weight);
        } catch (PersistenceException e) {
            throw new RuntimeException(e);
        }
    }

    private String getEdgeId(final String sourceNodeId, final String targetNodeId) {
        return sourceNodeId + " -> " + targetNodeId;
    }

    private PersistentObject getCreateNode(final String sourceNodeId, final Partition partition)
            throws PersistenceException {
        PersistentObject node = m_net.getNode(sourceNodeId);
        if (node == null) {
            node = m_net.createNode(sourceNodeId, partition);
        }
        return node;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public KPartiteGraphView<PersistentObject, Partition> getData() {
        return m_net;
    }
}