/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 13.05.2013 by koetter
 */
package edu.cmu.koetter.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.network.core.api.GraphObjectIterator;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.util.GraphUtil;

/**
 *
 * @author Tobias Koetter
 */
public class NeighborBiGraphExtractor {

    private Map<String, Set<String>> m_aspectNeighbors;
    private Map<String, Set<String>> m_memberNeighbors;
    private final KPartiteGraphView<PersistentObject, Partition> m_view;
    private final Map<String, NodeNeighbors> m_neighborMap;

    /**
     * @param view the {@link KPartiteGraphView}
     * @param neighborMap {@link Map} with node {@link NodeNeighbors}s
     */
    public NeighborBiGraphExtractor(final KPartiteGraphView<PersistentObject, Partition> view,
        final Map<String, NodeNeighbors> neighborMap) {
        m_view = view;
        m_neighborMap = neighborMap;
    }

    /**
     * @return the view
     */
    public KPartiteGraphView<PersistentObject, Partition> getView() {
        return m_view;
    }


    /**
     * Starts the conversion.
     * @param exec {@link ExecutionMonitor} to provide progress
     * @param minNodeDegree minimum node degree
     * @param startNode the start node
     * @throws PersistenceException if an exception occurs
     * @throws CanceledExecutionException if the operation has been canceled
     */
    public void run(final ExecutionMonitor exec, final int minNodeDegree,
                 final PersistentObject startNode) throws PersistenceException, CanceledExecutionException {
        exec.checkCanceled();
        exec.setMessage("start analysis of node " + startNode.getId());
        final Set<String> potentialMembers = m_neighborMap.get(startNode.getId()).getPredecessors();
        final int minAspectDegree = Math.max(potentialMembers.size(), minNodeDegree);
        m_memberNeighbors = new HashMap<String, Set<String>>(potentialMembers.size());
        m_aspectNeighbors = new HashMap<String, Set<String>>();
        for (final String pMember : potentialMembers) {
            final Set<String> potentialAspects = m_neighborMap.get(pMember).getSuccessors();
            if (potentialAspects.size() < minNodeDegree) {
                //skip all members that do not have enough aspects
                continue;
            }
            m_memberNeighbors.put(pMember, potentialAspects);
            for (final String pAspect : potentialAspects) {
                final Set<String> aMember = m_neighborMap.get(pAspect).getPredecessors();
                if (aMember.size() < minAspectDegree || m_aspectNeighbors.containsKey(aMember)) {
                    //skip all aspects that are more specific than the start node or that have already been processed
                    continue;
                }
                /*
                if (potentialMembers.contains(pAspect)) {
                    //skip all members from the aspects
//or check if they are connected to half of the other members?
                    continue;
                }*/
                m_aspectNeighbors.put(pAspect, aMember);
            }
        }
        exec.setProgress(1.0, startNode.getId() + " analyzed");
    }

    /**
     * @return the aspects and their neighbors as unmodifiable map
     */
    public Map<String, Set<String>> getAspectNeighbors() {
        return Collections.unmodifiableMap(m_aspectNeighbors);
    }

    /**
     * @return the members and their neighbors as unmodifiable map
     */
    public Map<String, Set<String>> getMemberNeighbors() {
        return Collections.unmodifiableMap(m_memberNeighbors);
    }

    /**
     * @param edgePartitions the node partitions
     * @param nodePartitions the edge partitions
     * @param node the id of the node to get the neighbors for
     * @param isSource <code>true</code> if the node should be the source and its target neighbors should be extracted
     * @return the set of neighbors
     */
    protected Set<String> getAdjacentNodeSet(final Collection<Partition> nodePartitions,
                                           final Collection<Partition> edgePartitions,
                                           final PersistentObject node, final boolean isSource) {
        GraphObjectIterator<PersistentObject> neighbors = null;
        GraphObjectIterator<PersistentObject> incidentEdges = null;
        final HashSet<String> set = new HashSet<String>();
        try {
            incidentEdges = getView().getIncidentObjects(node, edgePartitions);
            for (final PersistentObject edge : incidentEdges) {
                if (isSource && !getView().isSource(edge, node) || !isSource && !getView().isTarget(edge, node)) {
                    continue;
                }
                neighbors = getView().getIncidentObjects(edge, nodePartitions);
                for (final PersistentObject neighbor : neighbors) {
                    if (node.equals(neighbor)) {
                        continue;
                    }
                    if (isSource && !getView().isTarget(edge, neighbor)
                            || !isSource && !getView().isSource(edge, neighbor)) {
                        continue;
                    }
                    set.add(neighbor.getId());
                }
            }
            return set;
        } catch (final PersistenceException e) {
            throw new RuntimeException(e);
        } finally {
            GraphUtil.close(neighbors);
            GraphUtil.close(incidentEdges);
        }
    }
}
