/*
 * ------------------------------------------------------------------------
 *
 *  Copyright (C) 2003 - 2013
 *  University of Konstanz, Germany and
 *  KNIME GmbH, Konstanz, Germany
 *  Website: http://www.knime.org; Email: contact@knime.org
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License, Version 3, as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses>.
 *
 *  Additional permission under GNU GPL version 3 section 7:
 *
 *  KNIME interoperates with ECLIPSE solely via ECLIPSE's plug-in APIs.
 *  Hence, KNIME and ECLIPSE are both independent programs and are not
 *  derived from each other. Should, however, the interpretation of the
 *  GNU GPL Version 3 ("License") under any applicable laws result in
 *  KNIME and ECLIPSE being a combined program, KNIME GMBH herewith grants
 *  you the additional permission to use and propagate KNIME together with
 *  ECLIPSE with only the license terms in place for ECLIPSE applying to
 *  ECLIPSE and the GNU GPL Version 3 applying for KNIME, provided the
 *  license terms of ECLIPSE themselves allow for the respective use and
 *  propagation of ECLIPSE together with KNIME.
 *
 *  Additional permission relating to nodes for KNIME that extend the Node
 *  Extension (and in particular that are based on subclasses of NodeModel,
 *  NodeDialog, and NodeView) and that only interoperate with KNIME through
 *  standard APIs ("Nodes"):
 *  Nodes are deemed to be separate and independent programs and to not be
 *  covered works.  Notwithstanding anything to the contrary in the
 *  License, the License does not apply to Nodes, you are not required to
 *  license Nodes under the License, and you are granted a license to
 *  prepare and propagate Nodes, in each case even if such Nodes are
 *  propagated with or for interoperation with KNIME.  The owner of a Node
 *  may freely choose the license terms applicable to such Node, including
 *  when such Node is propagated with or for interoperation with KNIME.
 * ---------------------------------------------------------------------
 *
 * Created on 26.08.2013 by koetter
 */
package edu.cmu.koetter.util;


import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.Semaphore;
import java.util.concurrent.atomic.AtomicInteger;
import org.knime.core.node.CanceledExecutionException;
import org.knime.core.node.ExecutionMonitor;
import org.knime.core.node.KNIMEConstants;
import org.knime.core.util.ThreadPool;
import org.knime.network.core.api.GraphObjectIterator;
import org.knime.network.core.api.KPartiteGraphView;
import org.knime.network.core.api.Partition;
import org.knime.network.core.api.PersistentObject;
import org.knime.network.core.core.EndObject;
import org.knime.network.core.core.PartitionType;
import org.knime.network.core.core.exception.InvalidFeatureException;
import org.knime.network.core.core.exception.PersistenceException;
import org.knime.network.core.core.feature.StandardFeature;
import org.knime.network.core.core.util.GraphUtil;

/**
 *
 * @author koetter
 */
public class NeighborExtractor {

    private final Partition[] m_nodePartitions;
    private final Partition[] m_edgePartitions;
    private final KPartiteGraphView<PersistentObject, Partition> m_view;
    private final boolean m_directed;

    /**
     *
     * @param view the {@link KPartiteGraphView} to use
     * @throws PersistenceException if an exception in the network persistence layer occurs
     */
    public NeighborExtractor(final KPartiteGraphView<PersistentObject, Partition> view) throws PersistenceException {
        m_view = view;
        m_directed = m_view.isDirected();
        m_nodePartitions = m_view.getPartitions(PartitionType.NODE).toArray(new Partition[0]);
        m_edgePartitions = m_view.getPartitions(PartitionType.EDGE).toArray(new Partition[0]);
    }

    /**
     * @param exec {@link ExecutionMonitor} to provide progress
     * @param procCount number of processes to use
     * @return {@link Map} with the node neighbors
     * @throws InterruptedException if a thread has been interrupted
     * @throws PersistenceException if an exception in the network persistence layer occurs
     * @throws CanceledExecutionException if the user has canceled the operation
     */
    public Map<String, NodeNeighbors> getNodeNeighbors(final ExecutionMonitor exec, final int procCount)
                throws InterruptedException, PersistenceException, CanceledExecutionException {
        final long noOfNodes = m_view.getNoOfNodes();
        final HashMap<String, NodeNeighbors> neighborMap = new HashMap<String, NodeNeighbors>((int)noOfNodes);
        final ThreadPool pool =
                KNIMEConstants.GLOBAL_THREAD_POOL.createSubPool();
        //The semaphore restricts the number of concurrent processes
        final Semaphore semaphore = new Semaphore(procCount);
        if (procCount == 1) {
            //acquire the only permit to prevent memory problems
            semaphore.acquire();
        }
        GraphObjectIterator<PersistentObject> nodes = null;
        try {
            nodes = m_view.getNodes(m_nodePartitions);
            final AtomicInteger nodeCounter = new AtomicInteger();
            for (final PersistentObject node : nodes) {
                exec.checkCanceled();
                pool.enqueue(getNeighbors(exec, nodeCounter, noOfNodes, neighborMap, node));
            }
            if (procCount == 1) {
                //release the only permit to start the Runnables
                semaphore.release();
            }
            pool.waitForTermination();
            return neighborMap;
        } finally {
            GraphUtil.close(nodes);
        }
    }

    private Runnable getNeighbors(final ExecutionMonitor exec,
        final AtomicInteger counter, final long noOfNodes, final HashMap<String, NodeNeighbors> neighborMap,
        final PersistentObject node) {
        return new Runnable() {
            @Override
            public void run() {
                GraphObjectIterator<PersistentObject> edges = null;
                GraphObjectIterator<PersistentObject> neighbors = null;
                try {
                    exec.checkCanceled();
                    final HashSet<String> successors = new HashSet<String>();
                    final HashSet<String> predecessors;
                    if (m_directed) {
                        predecessors = new HashSet<String>();
                    } else {
                        predecessors = successors;
                    }
                    edges = m_view.getIncidentEdges(node, m_edgePartitions);
                    for (final PersistentObject edge : edges) {
                        exec.checkCanceled();
                        final boolean isTarget = !m_directed || isTarget(m_view, edge, node);
                        neighbors = m_view.getIncidentNodes(edge, m_nodePartitions);
                        for (final PersistentObject neighbor : neighbors) {
                            final String neighborId = neighbor.getId();
                            if (neighbor.equals(node)
                                    || (!m_directed && (successors.contains(neighborId)
                                            || predecessors.contains(neighborId)))) {
                                //skip the start node and all processed neighbors if the network is undirected
                                continue;
                            }
                            final boolean neighborIsTarget =
                                    !m_directed || isTarget(m_view, edge, neighbor);
                            if (m_directed) {
                                if (isTarget == neighborIsTarget) {
                                    //skip all neighbors that have the same state like the original node
                                    continue;
                                }
                                if (isTarget) {
                                    //the node is the target of the edge which makes it neighbor the source of the edge
                                    //hence the nodes predecessor
                                    predecessors.add(neighborId);
                                } else {
                                    successors.add(neighborId);
                                }
                            } else {
                                //the network and its edges are not directed so we have to add the neighbor only to the
                                //successor set since it is also used as predecessors
                                successors.add(neighborId);
                            }
                        }
                    }
                    final NodeNeighbors neighbor = new NodeNeighbors(m_directed, predecessors, successors);
                    synchronized (neighborMap) {
                        //add the neighbors to the map in a synchronized block
                        neighborMap.put(node.getId(), neighbor);
                    }
                    final int count = counter.incrementAndGet();
                    exec.setProgress(count / (double) noOfNodes,
                        "Extracting neighbors for node " + count + " of " + noOfNodes);
                } catch (final CanceledExecutionException e) {
                    // that is fine they are handled outside
                } catch (final Exception e) {
                    throw new RuntimeException(e);
                } finally {
                    GraphUtil.close(edges);
                    GraphUtil.close(neighbors);
                }
            }

            private boolean isTarget(
                    final KPartiteGraphView<PersistentObject, Partition> view,
                    final PersistentObject edge, final PersistentObject node) throws InvalidFeatureException, PersistenceException {
                final Boolean target = view.getBooleanFeature(new EndObject(node, edge),
                        StandardFeature.IS_TARGET);
                return target != null && target.booleanValue();
            }
        };
    }
}
